"""
Utilities for generating an :mod:`argparse` based CLI.

.. versionadded:: 2.0.0
"""

import argparse
import itertools
from datetime import date, datetime, timedelta
from enum import Enum
from functools import partial, wraps
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)


if TYPE_CHECKING:
    from argparse import FileType

from . import _core, converters
from .cli_utils import (
    NO_DEFAULT,
    Default,
    DefaultFactorySentinel,
    StrDict,
    TypeArgsMaker,
    TypeHandlerFunc,
    get_default,
)
from .constants import ARGPARSE_METADATA_KEY as METADATA_KEY
from .converters import Converter
from .loaders import Loader
from .processors import Processor
from .types import (
    SECRET_REPR,
    ST,
    LoadedValue,
    LoaderMeta,
    MergedSettings,
    OptionInfo,
)


__all__ = [
    "cli",
    "make_parser",
    "namespace2settings",
    "handle_datetime",
    "handle_enum",
    "handle_path",
    "DEFAULT_TYPES",
    "ArgparseHandler",
    "BooleanOptionalAction",
    "ListAction",
    "DictItemAction",
]


WrapppedFunc = Callable[[ST], Any]
CliFn = Callable[[ST], Any]
DecoratedCliFn = Callable[[], Optional[int]]


def handle_datetime(type: type, default: Default, is_optional: bool) -> StrDict:
    """
    Handle isoformatted datetimes.
    """
    kwargs: StrDict = {
        "type": partial(converters.to_datetime, cls=datetime),
        "metavar": "YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
    }
    if isinstance(default, datetime):
        kwargs["default"] = default.isoformat()
    elif is_optional:
        kwargs["default"] = None
    return kwargs


def handle_date(type: type, default: Default, is_optional: bool) -> StrDict:
    """
    Handle isoformatted datetimes.
    """
    kwargs: StrDict = {
        "type": partial(converters.to_date, cls=date),
        "metavar": "YYYY-MM-DD",
    }
    if isinstance(default, date):
        kwargs["default"] = default.isoformat()
    elif is_optional:
        kwargs["default"] = None
    return kwargs


def handle_timedelta(type: type, default: Default, is_optional: bool) -> StrDict:
    """
    Handle isoformatted datetimes.
    """
    kwargs: StrDict = {
        "type": partial(converters.to_timedelta, cls=timedelta),
        "metavar": "[-][Dd][HHh][MMm][SS[.ffffff]s]",
    }
    if isinstance(default, timedelta):
        kwargs["default"] = converters.timedelta_to_str(default)
    elif is_optional:
        kwargs["default"] = None
    return kwargs


def handle_enum(type: Type[Enum], default: Default, is_optional: bool) -> StrDict:
    """
    Use *choices* as option type and use the enum value's name as default.
    """
    kwargs: StrDict = {"choices": list(type.__members__)}
    if isinstance(default, type):
        # Convert Enum instance to string
        kwargs["default"] = default.name
    elif is_optional:
        kwargs["default"] = None

    return kwargs


def handle_path(type: Type[Path], default: Default, is_optional: bool) -> StrDict:
    """
    Handle :class:`pathlib.Path` and also use proper metavar.
    """
    kwargs: StrDict = {"type": Path, "metavar": "PATH"}
    if isinstance(default, (Path, str)):
        kwargs["default"] = str(default)
    elif is_optional:
        kwargs["default"] = None

    return kwargs


#: Default handlers for argparse option types.
DEFAULT_TYPES: Dict[type, TypeHandlerFunc] = {
    datetime: handle_datetime,
    date: handle_date,
    timedelta: handle_timedelta,
    Enum: handle_enum,
    Path: handle_path,
}


class ArgparseHandler:
    """
    Implementation of the :class:`~typed_settings.cli_utils.TypeHandler`
    protocol for Click.

    Args:
        extra_types: A dict mapping types to specialized handler functions.
            Use :data:`DEFAULT_TYPES` by default.
    """

    def __init__(
        self, extra_types: Optional[Dict[type, TypeHandlerFunc]] = None
    ) -> None:
        self.extra_types = extra_types or DEFAULT_TYPES

    def get_scalar_handlers(self) -> Dict[type, TypeHandlerFunc]:
        return self.extra_types

    def handle_scalar(
        self,
        type: Optional[type],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        kwargs: StrDict = {"type": type}
        if type is not None:
            if issubclass(type, str):
                kwargs["metavar"] = "TEXT"
            else:
                kwargs["metavar"] = str(type.__name__).upper()

        # if default is not None or is_optional:
        if default not in (None, NO_DEFAULT):
            kwargs["default"] = default
        elif is_optional:
            kwargs["default"] = None
        if type and issubclass(type, bool):
            kwargs["action"] = BooleanOptionalAction

        return kwargs

    def handle_tuple(
        self,
        type_args_maker: TypeArgsMaker,
        types: Tuple[Any, ...],
        default: Optional[Tuple],
        is_optional: bool,
    ) -> StrDict:
        metavar = tuple(
            "TEXT" if issubclass(t, str) else str(t.__name__).upper() for t in types
        )
        kwargs = {
            "metavar": metavar,
            "nargs": len(types),
            "default": default,
        }
        return kwargs

    def handle_collection(
        self,
        type_args_maker: TypeArgsMaker,
        types: Tuple[Any, ...],
        default: Optional[Collection[Any]],
        is_optional: bool,
    ) -> StrDict:
        kwargs = type_args_maker.get_kwargs(types[0], NO_DEFAULT)
        kwargs["default"] = default or []  # Don't use None as default
        kwargs["action"] = ListAction
        return kwargs

    def handle_mapping(
        self,
        type_args_maker: TypeArgsMaker,
        types: Tuple[Any, ...],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        kwargs = {
            "metavar": "KEY=VALUE",
            "action": DictItemAction,
        }
        if not isinstance(default, Mapping):
            default = {}
        kwargs["default"] = default
        kwargs["default_repr"] = ", ".join(f"{k}={v}" for k, v in default.items())

        return kwargs


def cli(
    settings_cls: Type[ST],
    loaders: Union[str, Sequence[Loader]],
    *,
    processors: Sequence[Processor] = (),
    converter: Optional[Converter] = None,
    base_dir: Path = Path(),
    type_args_maker: Optional[TypeArgsMaker] = None,
    **parser_kwargs: Any,
) -> Callable[[CliFn[ST]], DecoratedCliFn]:
    r"""
    **Decorator:** Generate an argument parser for the options of the given
    settings class and pass an instance of that class to the decorated
    function.

    Args:
        settings_cls: The settings class to generate options for.

        loaders: Either a string with your app name or a list of
            :class:`.Loader`\ s.  If it's a string, use it with
            :func:`~typed_settings.default_loaders()` to get the default
            loaders.

        processors: A list of settings :class:`.Processor`'s.

        converter: An optional :class:`.Converter` used for converting option values to
            the required type.

            By default, :data:`typed_settings.default_converter()` is used.

        base_dir: Base directory for resolving relative paths in default option values.

        type_args_maker: The type args maker that is used to generate keyword
            arguments for :func:`click.option()`.  By default, use
            :class:`.TypeArgsMaker` with :class:`ArgparseHandler`.

        **parser_kwargs: Additional keyword arguments to pass to the
            :class:`argparse.ArgumentParser`.

    Return:
        A decorator for an argparse CLI function.

    Raise:
        InvalidSettingsError: If an instance of *cls* cannot be created for the given
            settings.

    Example:
        .. code-block:: python

           import typed_settings as ts

           @ts.settings
           class Settings: ...

           @ts.cli(Settings, "example")
           def cli(settings: Settings) -> None:
               print(settings)

    .. versionchanged:: 23.0.0
       Made *converter* and *type_args_maker* a keyword-only argument
    .. versionchanged:: 23.0.0
       Added the *processors* argument
    .. versionchanged:: 23.1.0
       Added the *base_dir* argument
    """
    if isinstance(loaders, str):
        loaders = _core.default_loaders(loaders)
    converter = converter or converters.default_converter()
    state = _core.SettingsState(settings_cls, loaders, processors, converter, base_dir)
    type_args_maker = type_args_maker or TypeArgsMaker(ArgparseHandler())

    decorator = _get_decorator(state, type_args_maker, **parser_kwargs)
    return decorator


def make_parser(
    settings_cls: Type[ST],
    loaders: Union[str, Sequence[Loader]],
    *,
    processors: Sequence[Processor] = (),
    converter: Optional[Converter] = None,
    base_dir: Path = Path(),
    type_args_maker: Optional[TypeArgsMaker] = None,
    **parser_kwargs: Any,
) -> Tuple[argparse.ArgumentParser, MergedSettings]:
    r"""
    Return an argument parser for the options of the given settings class.

    Use :func:`namespace2settings()` to convert the parser's namespace to an
    instance of the settings class.

    Args:
        settings_cls: The settings class to generate options for.

        loaders: Either a string with your app name or a list of
            :class:`.Loader`\ s.  If it's a string, use it with
            :func:`~typed_settings.default_loaders()` to get the default
            loaders.

        processors: A list of settings :class:`.Processor`'s.

        converter: An optional :class:`.Converter` used for converting option values to
            the required type.

            By default, :data:`typed_settings.default_converter()` is used.

        base_dir: Base directory for resolving relative paths in default option values.

        type_args_maker: The type args maker that is used to generate keyword
            arguments for :func:`click.option()`.  By default, use
            :class:`.TypeArgsMaker` with :class:`ArgparseHandler`.

        **parser_kwargs: Additional keyword arguments to pass to the
            :class:`argparse.ArgumentParser`.

    Return:
        An argument parser configured with with an argument for each option of
        *settings_cls*.

    Raise:
        InvalidSettingsError: If an instance of *cls* cannot be created for the given
            settings.

    .. versionchanged:: 23.0.0
       Made *converter* and *type_args_maker* a keyword-only argument
    .. versionchanged:: 23.0.0
       Added the *processors* argument
    .. versionchanged:: 23.1.0
       Added the *base_dir* argument
    """
    if isinstance(loaders, str):
        loaders = _core.default_loaders(loaders)
    converter = converter or converters.default_converter()
    state = _core.SettingsState(settings_cls, loaders, processors, converter, base_dir)
    type_args_maker = type_args_maker or TypeArgsMaker(ArgparseHandler())

    return _mk_parser(state, type_args_maker, **parser_kwargs)


def namespace2settings(
    settings_cls: Type[ST],
    namespace: argparse.Namespace,
    *,
    merged_settings: MergedSettings,
    converter: Optional[Converter] = None,
    base_dir: Path = Path(),
) -> ST:
    """
    Create a settings instance from an argparse namespace.

    To be used together with :func:`make_parser()`.

    Args:
        settings_cls: The settings class to instantiate.
        namespace: The namespace returned by the argument parser.
        merged_settings: The loaded and merged settings by settings name.
        converter: An optional :class:`.Converter` used for converting option values to
            the required type.  By default, :data:`typed_settings.default_converter()`
            is used.
        base_dir: Base directory for resolving relative paths in default option values.

    Raise:
        InvalidSettingsError: If an instance of *cls* cannot be created for the given
            settings.

    Return: An instance of *settings_cls*.

    .. versionchanged:: 23.1.0
       Added the *base_dir* argument
    """
    converter = converter or converters.default_converter()
    state = _core.SettingsState(settings_cls, [], [], converter, base_dir)
    return _ns2settings(namespace, state, merged_settings)


def _get_decorator(
    state: _core.SettingsState[ST],
    type_args_maker: TypeArgsMaker,
    **parser_kwargs: Any,
) -> Callable[[CliFn], DecoratedCliFn]:
    """
    Build the CLI decorator based on the user's config.
    """

    def decorator(func: CliFn) -> DecoratedCliFn:
        """
        Create an argument parsing wrapper for *func*.

        The wrapper

        - loads settings as default option values
        - creates an argument parser with an option for each setting
        - parses the command line options
        - passes the updated settings instance to the decorated function
        """

        @wraps(func)
        def cli_wrapper() -> Optional[int]:
            if "description" not in parser_kwargs and func.__doc__:
                parser_kwargs["description"] = func.__doc__.strip()
            parser, merged_settings = _mk_parser(
                state, type_args_maker, **parser_kwargs
            )

            args = parser.parse_args()
            settings = _ns2settings(args, state, merged_settings)
            return func(settings)

        return cli_wrapper

    return decorator


def _mk_parser(
    state: _core.SettingsState[ST],
    type_args_maker: TypeArgsMaker,
    **parser_kwargs: Any,
) -> Tuple[argparse.ArgumentParser, MergedSettings]:
    """
    Create an :class:`argparse.ArgumentParser` for all options.
    """
    merged_settings = _core._load_settings(state)
    grouped_options = [
        (g_cls, list(g_opts))
        for g_cls, g_opts in itertools.groupby(
            state.options, key=lambda o: o.parent_cls
        )
    ]
    parser = argparse.ArgumentParser(**parser_kwargs)
    for g_cls, g_opts in grouped_options:
        group = parser.add_argument_group(g_cls.__name__, f"{g_cls.__name__} options")
        for oinfo in g_opts:
            default = get_default(oinfo, merged_settings, state.converter)
            flags, cfg = _mk_argument(oinfo, default, type_args_maker)
            group.add_argument(*flags, **cfg)
    return (parser, merged_settings)


def _mk_argument(
    oinfo: OptionInfo,
    default: Default,
    type_args_maker: TypeArgsMaker,
) -> Tuple[Tuple[str, ...], Dict[str, Any]]:
    user_config = dict(oinfo.metadata.get(METADATA_KEY, {}))

    # The option type specifies the default option kwargs
    kwargs = type_args_maker.get_kwargs(oinfo.cls, default)

    param_decls: Tuple[str, ...]
    user_param_decls: Union[str, Sequence[str]]
    user_param_decls = user_config.pop("param_decls", ())
    if not user_param_decls:
        option_name = oinfo.path.replace(".", "-").replace("_", "-")
        param_decls = (f"--{option_name}",)
    elif isinstance(user_param_decls, str):
        param_decls = (user_param_decls,)
    else:
        param_decls = tuple(user_param_decls)

    # Get "help" from the user_config *now*, because we may need to update it
    # below.  Also replace "None" with "".
    kwargs["help"] = user_config.pop("help", None) or ""
    if "default" in kwargs and kwargs["default"] is not NO_DEFAULT:
        default_repr = kwargs.pop("default_repr", kwargs["default"])
        if kwargs["default"] is None:
            help_extra = ""
        elif oinfo.is_secret:
            help_extra = f" [default: ({SECRET_REPR})]"
        elif isinstance(callable(kwargs["default"]), DefaultFactorySentinel):
            help_extra = "[default: (dynamic)]"
        else:
            help_extra = f" [default: {default_repr}]"
        if kwargs["default"] not in (None, (), [], {}):
            kwargs["default"] = DefaultFactorySentinel()
    else:
        kwargs["required"] = True
        help_extra = " [required]"
    kwargs["help"] = f"{kwargs['help']}{help_extra}"

    # The user has the last word, though.
    kwargs.update(user_config)

    return (param_decls, kwargs)


def _ns2settings(
    namespace: argparse.Namespace,
    state: _core.SettingsState[ST],
    merged_settings: MergedSettings,
) -> ST:
    """
    Convert the :class:`argparse.Namespace` to an instance of the settings
    class and return it.
    """
    meta = LoaderMeta("Command line args")
    for option_info in state.options:
        path = option_info.path
        attr = path.replace(".", "_")
        if hasattr(namespace, attr):  # pragma: no cover
            # "path" *should* always be in "cli_options", b/c we *currently*
            # generate CLI options for all options.  But let's stay safe here
            # in case the behavior changes in the future.
            value = getattr(namespace, attr)
            if not isinstance(value, DefaultFactorySentinel):
                merged_settings[path] = LoadedValue(value, meta)
    settings = _core.convert(merged_settings, state)
    return settings


class BooleanOptionalAction(argparse.Action):
    """
    An argparse action for handling boolean flags.
    """

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        default: Default = None,
        type: Union[Callable[[str], Any], "FileType", None] = None,
        choices: Optional[Iterable[Any]] = None,
        required: bool = False,
        help: Optional[str] = None,
        metavar: Union[str, Tuple[str, ...], None] = None,
    ) -> None:
        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if not option_string.startswith("--"):
                raise ValueError(
                    f"Only boolean flags starting with '--' are supported: "
                    f"{option_string}"
                )
            option_string = "--no-" + option_string[2:]
            _option_strings.append(option_string)

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        if option_string and option_string in self.option_strings:  # pragma: no cover
            setattr(namespace, self.dest, not option_string.startswith("--no-"))

    def format_usage(self) -> str:
        return " | ".join(self.option_strings)


class ListAction(argparse.Action):
    """
    An argparse action for handling lists.
    """

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        nargs: Union[int, str, None] = None,
        default: Default = None,
        type: Union[Callable[[str], Any], "FileType", None] = None,
        choices: Optional[Iterable[Any]] = None,
        required: bool = False,
        help: Optional[str] = None,
        metavar: Union[str, Tuple[str, ...], None] = None,
    ) -> None:
        if nargs == 0:  # pragma: no cover
            raise ValueError(f"nargs for append actions must be != 0: {nargs}")
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        if values is None:
            return  # pragma: no cover

        items = getattr(namespace, self.dest, [])
        # Do not append to the defaults but create a new list!
        if items is self.default:
            items = []
        items.append(values)
        setattr(namespace, self.dest, items)


class DictItemAction(argparse.Action):
    """
    An argparse action for handling dicts.
    """

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        default: Default = None,
        type: Union[Callable[[str], Any], "FileType", None] = None,
        choices: Optional[Iterable[Any]] = None,
        required: bool = False,
        help: Optional[str] = None,
        metavar: Union[str, Tuple[str, ...], None] = None,
    ) -> None:
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=1,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        if values is None:
            return  # pragma: no cover

        if isinstance(values, str):
            values = [values]  # pragma: no cover

        items = getattr(namespace, self.dest, {})
        # Do not append to the defaults but create a new list!
        if items is self.default:
            items = {}

        for value in values:
            k, _, v = value.partition("=")
            items[k] = v

        setattr(namespace, self.dest, items)



"""
Core functions for loading and working with settings.
"""

from typing import Any, List

from ._core import SettingsState, convert, default_loaders, load, load_settings
from ._file_utils import find
from .cli_argparse import cli
from .cls_utils import resolve_types
from .converters import default_converter, register_strlist_hook
from .loaders import EnvLoader, FileLoader, TomlFormat
from .types import Secret, SecretStr


_attrs_imports = {"combine", "evolve", "option", "secret", "settings"}
_click_imports = {"click_options", "pass_settings"}

try:
    from .cls_attrs import combine, evolve, option, secret, settings
except ImportError:
    pass

try:
    from .cli_click import click_options, pass_settings
except ImportError:
    pass


def __getattr__(name: str) -> Any:
    """
    Raise a helpful :exc:`ModuleNotFound` error when getting something that requires an
    optional dependency that is not installed.
    """
    # This method is only invoked if either
    # - attrs/click is not installed or
    # - an attribute that actually doesn't exist
    # is requested.
    if name in _attrs_imports:
        raise ModuleNotFoundError(
            "Module 'attrs' not installed.  Please run "
            "'python -m pip install -U typed-settings[attrs]'"
        )

    if name in _click_imports:
        raise ModuleNotFoundError(
            "Module 'click' not installed.  Please run "
            "'python -m pip install -U typed-settings[click]'"
        )

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    # Core
    "SettingsState",
    "default_loaders",
    "load",
    "load_settings",
    "convert",
    # Types
    "Secret",
    "SecretStr",
    # Class utils
    "resolve_types",
    # File utils
    "find",
    # Loaders
    "EnvLoader",
    "FileLoader",
    "TomlFormat",
    # Cattrs converters/helpers
    "default_converter",
    "register_strlist_hook",
    # Argparse utils
    "cli",
    # Optional: attrs
    "combine",
    "evolve",
    "option",
    "secret",
    "settings",
    # Optional: click
    "click_options",
    "pass_settings",
]


def __dir__() -> List[str]:
    return __all__



"""
Tests for "typed_settings.cli.click".
"""

import unittest.mock as mock
from pathlib import Path
from typing import Any, Callable, Generic, List, Optional, TypeVar, Union

import attrs
import click
import click.testing
import pydantic
import pytest

import typed_settings.cli_click as cli_click
from typed_settings import (
    click_options,
    default_loaders,
    option,
    pass_settings,
    secret,
    settings,
)
from typed_settings.constants import METADATA_KEY
from typed_settings.types import SecretStr


T = TypeVar("T")


Invoke = Callable[..., click.testing.Result]


class CliResult(click.testing.Result, Generic[T]):
    """A container for the settings passed to a test CLI."""

    settings: Optional[T]


Cli = Callable[..., CliResult[T]]


@pytest.fixture(name="invoke")
def _invoke() -> Invoke:
    runner = click.testing.CliRunner()

    def invoke(cli: click.Command, *args: str) -> click.testing.Result:
        return runner.invoke(cli, args, catch_exceptions=False)

    return invoke


def test_simple_cli(invoke: Invoke) -> None:
    """
    Basic test "click_options()", create a simple CLI for simple settings.
    """

    @settings
    class Settings:
        o: int

    @click.command()
    @click_options(Settings, "test")
    def cli(settings: Settings) -> None:
        assert settings == Settings(3)

    invoke(cli, "--o=3")


def test_unkown_type(invoke: Invoke) -> None:
    """
    A TypeError is raised if the settings contain a type that the decorator
    cannot handle.
    """

    @settings
    class Settings:
        o: Union[int, str]

    with pytest.raises(
        TypeError,
        match=r"Cannot create CLI option for: typing.Union\[int, str\]",
    ):

        @click.command()  # pragma: no cover
        @click_options(Settings, "test")
        def cli(settings: Settings) -> None: ...


def test_attrs_meta_not_modified() -> None:
    """
    The attrs meta data with with user defined click config is not modified.

    Regression test for #29.
    """

    @settings
    class S:
        opt: int = option(help="spam", click={"callback": print})

    meta = attrs.fields(S).opt.metadata[METADATA_KEY]

    assert meta[cli_click.METADATA_KEY] == {"help": "spam", "callback": print}

    click_options(S, "test")(lambda s: None)  # pragma: no cover
    click_options(S, "test")(lambda s: None)  # pragma: no cover

    assert meta[cli_click.METADATA_KEY] == {"help": "spam", "callback": print}


class TestDefaultsLoading:
    """
    Tests for loading default values.
    """

    def test_no_default(self, invoke: Invoke, monkeypatch: pytest.MonkeyPatch) -> None:
        """
        cli_options without a default are mandatory/required.
        """

        @settings
        class Settings:
            a: str
            b: str

        monkeypatch.setenv("TEST_A", "spam")  # This makes only "S.b" mandatory!

        @click.command()
        @click_options(Settings, default_loaders("test"))
        def cli(settings: Settings) -> None: ...

        result = invoke(cli)
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "Try 'cli --help' for help.\n"
            "\n"
            "Error: Missing option '--b'.\n"
        )
        assert result.exit_code == 2

    def test_help_text(self, invoke: Invoke) -> None:
        """
        cli_options/secrets can specify a help text for click cli_options.
        """

        @settings
        class Settings:
            a: str = option(default="spam", help="Help for 'a'")
            b: str = secret(default="eggs", help="bbb")

        @click.command()
        @click_options(Settings, default_loaders("test"))
        def cli(settings: Settings) -> None: ...

        result = invoke(cli, "--help")
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --a TEXT  Help for 'a'  [default: spam]\n"
            "  --b TEXT  bbb  [default: (*******)]\n"
            "  --help    Show this message and exit.\n"
        )
        assert result.exit_code == 0

    def test_help_text_secrets(self, invoke: Invoke) -> None:
        """
        Defaults for Secret(Str) type defaults are not show even if "option()"
        and not "secret()" is used.
        """

        @settings
        class Settings:
            a: SecretStr = SecretStr("spam")
            # b: Secret[int] = Secret(42)

        @click.command()
        @click_options(Settings, default_loaders("test"))
        def cli(settings: Settings) -> None: ...

        result = invoke(cli, "--help")
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --a TEXT  [default: (*******)]\n"
            "  --help    Show this message and exit.\n"
        )
        assert result.exit_code == 0

    def test_show_envvar_not_in_help(self, invoke: Invoke) -> None:
        """
        The env var will not be shown if the envloader is not being used.
        """

        @settings
        class Settings:
            a: str = "spam"
            b: str = secret(default="eggs")

        @click.command()
        @click_options(Settings, [], show_envvars_in_help=True)
        def cli(settings: Settings) -> None: ...

        result = invoke(cli, "--help")
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --a TEXT  [default: spam]\n"
            "  --b TEXT  [default: (*******)]\n"
            "  --help    Show this message and exit.\n"
        )
        assert result.exit_code == 0

    def test_long_name(self, invoke: Invoke) -> None:
        """
        Underscores in option names are replaces with "-" in Click cli_options.
        """

        @settings
        class Settings:
            long_name: str = "val"

        @click.command()
        @click_options(Settings, default_loaders("test"))
        def cli(settings: Settings) -> None: ...

        result = invoke(cli, "--help")
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --long-name TEXT  [default: val]\n"
            "  --help            Show this message and exit.\n"
        )
        assert result.exit_code == 0

    def test_click_default_from_settings(
        self, invoke: Invoke, monkeypatch: pytest.MonkeyPatch, tmp_path: Path
    ) -> None:
        """
        If a setting is set in a config file, that value is being used as
        default for click cli_options - *not* the default defined in the
        Settings class.
        """
        tmp_path.joinpath("settings.toml").write_text('[test]\na = "x"\n')
        spath = tmp_path.joinpath("settings2.toml")
        spath.write_text('[test]\nb = "y"\n')
        monkeypatch.setenv("TEST_SETTINGS", str(spath))
        monkeypatch.setenv("TEST_C", "z")

        @settings
        class Settings:
            a: str
            b: str
            c: str
            d: str

        @click.command()
        @click_options(
            Settings,
            default_loaders("test", [tmp_path.joinpath("settings.toml")]),
        )
        def cli(settings: Settings) -> None: ...

        result = invoke(cli, "--help")
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --a TEXT  [default: x]\n"
            "  --b TEXT  [default: y]\n"
            "  --c TEXT  [default: z]\n"
            "  --d TEXT  [required]\n"
            "  --help    Show this message and exit.\n"
        )
        assert result.exit_code == 0

    def test_default_factory_multiple_invocation(self, invoke: Invoke) -> None:
        """
        Default factories are not invoked by click when the CLI is generated.
        They are evaluate during the "convert" phase each time the CLI is invoked.
        """
        loaded_settings: List["Settings"] = []  # noqa: UP037

        @settings
        class Settings:
            o: int = option(factory=lambda: len(loaded_settings) + 1)

        @click.command()
        @click_options(Settings, "example")
        def cli(settings: Settings) -> None:
            loaded_settings.append(settings)

        invoke(cli)
        invoke(cli)
        invoke(cli, "--o=100")
        assert loaded_settings == [Settings(1), Settings(2), Settings(100)]


class TestSettingsPassing:
    """
    Test for passing settings as positional or keyword arg.
    """

    def test_pass_as_pos_arg(self, invoke: Invoke) -> None:
        """
        If no explicit argname is provided, the settings instance is passed
        as positional argument.
        """

        @settings
        class Settings:
            o: int

        @click.command()
        @click_options(Settings, "test")
        # We should to this in this test:
        #   def cli(settings: Settings, /) -> None:
        # but that does not work in py37, so we just use name that is
        # != CTX_KEY.
        def cli(s: Settings) -> None:
            assert s == Settings(3)

        invoke(cli, "--o=3")

    def test_pos_arg_order_1(self, invoke: Invoke) -> None:
        """
        The inner most decorator maps to the first argument.
        """

        @settings
        class Settings:
            o: int = 0

        @click.command()
        @click_options(Settings, "test")
        @click.pass_obj
        # def cli(obj: dict, settings: Settings, /) -> None:
        def cli(obj: dict, settings: Settings) -> None:
            assert settings == Settings(3)
            assert obj["settings"] is settings

        result = invoke(cli, "--o=3")
        assert result.exit_code == 0, result.output

    def test_pos_arg_order_2(self, invoke: Invoke) -> None:
        """
        The inner most decorator maps to the first argument.

        Variant of "test_pos_arg_order_1" with swapeed decorators/args.
        """

        @settings
        class Settings:
            o: int = 0

        @click.command()
        @click.pass_obj
        @click_options(Settings, "test")
        # def cli(settings: Settings, obj: dict, /) -> None:
        def cli(settings: Settings, obj: dict) -> None:
            assert settings == Settings(3)
            assert obj["settings"] is settings

        result = invoke(cli, "--o=3")
        assert result.exit_code == 0, result.output

    def test_change_arg_name(self, invoke: Invoke) -> None:
        """
        The name of the settings argument can be changed.  It is then passed
        as kwarg.
        """

        @settings
        class Settings:
            o: int

        @click.command()
        @click_options(Settings, "test", argname="le_settings")
        def cli(*, le_settings: Settings) -> None:
            assert le_settings == Settings(3)

        result = invoke(cli, "--o=3")
        assert result.exit_code == 0, result.output

    def test_multi_settings(self, invoke: Invoke) -> None:
        """
        Multiple settings classes can be used when the argname is changed.
        """

        @settings
        class A:
            a: int = 0

        @settings
        class B:
            b: str = "b"

        @click.command()
        @click_options(A, "test-a", argname="sa")
        @click_options(B, "test-b", argname="sb")
        def cli(*, sa: A, sb: B) -> None:
            assert sa == A()
            assert sb == B()
            print("ok")

        result = invoke(cli)
        assert result.output == "ok\n"

        result = invoke(cli, "--help")
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --a INTEGER  [default: 0]\n"
            "  --b TEXT     [default: b]\n"
            "  --help       Show this message and exit.\n"
        )

    def test_multi_settings_duplicates(self, invoke: Invoke) -> None:
        """
        Different settings classes should not define the same options!
        """

        @settings
        class A:
            a: int = 0

        @settings
        class B:
            a: str = 3  # type: ignore
            b: str = "b"

        @click.command()
        @click_options(A, "test-a", argname="sa")
        @click_options(B, "test-b", argname="sb")
        def cli(*, sa: A, sb: B) -> None: ...

        result = invoke(cli, "--help")
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --a INTEGER  [default: 0]\n"
            "  --a TEXT     [default: 3]\n"
            "  --b TEXT     [default: b]\n"
            "  --help       Show this message and exit.\n"
        )

    def test_empty_cls(self, invoke: Invoke) -> None:
        """
        Empty settings classes are no special case.
        """

        @settings
        class S:
            pass

        @click.command()
        @click_options(S, "test")
        def cli(settings: S) -> None:
            assert settings == S()

        invoke(cli)


class TestPassSettings:
    """Tests for pass_settings()."""

    @settings
    class Settings:  # noqa: D106
        opt: str = ""

    def test_pass_settings(self, invoke: Invoke) -> None:
        """
        A subcommand can receive the settings (as pos arg) via the
        `pass_settings` decorator.
        """

        @click.group()
        @click_options(self.Settings, default_loaders("test"))
        def cli(settings: TestPassSettings.Settings) -> None:
            pass

        @cli.command()
        @pass_settings
        def cmd(s: TestPassSettings.Settings) -> None:
            assert s == self.Settings(opt="spam")

        invoke(cli, "--opt=spam", "cmd")

    def test_change_argname(self, invoke: Invoke) -> None:
        """
        The argument name for "pass_settings" can be changed but must be the
        same as in "click_options()".
        """

        @click.group()
        @click_options(self.Settings, "test", argname="le_settings")
        def cli(le_settings: TestPassSettings.Settings) -> None:
            pass

        @cli.command()
        @pass_settings(argname="le_settings")
        def cmd(*, le_settings: TestPassSettings.Settings) -> None:
            assert le_settings == self.Settings(opt="spam")

        invoke(cli, "--opt=spam", "cmd")

    def test_pass_settings_no_settings(self, invoke: Invoke) -> None:
        """
        Pass ``None`` if no settings are defined.
        """

        @click.group()
        def cli() -> None:
            pass

        @cli.command()
        @pass_settings
        def cmd(settings: TestPassSettings.Settings) -> None:
            assert settings is None

        invoke(cli, "cmd")

    def test_change_argname_no_settings(self, invoke: Invoke) -> None:
        """
        Pass ``None`` if no settings are defined.
        """

        @click.group()
        def cli() -> None:
            pass

        @cli.command()
        @pass_settings(argname="le_settings")
        def cmd(le_settings: TestPassSettings.Settings) -> None:
            assert le_settings is None

        invoke(cli, "cmd")

    def test_pass_in_parent_context(self, invoke: Invoke) -> None:
        """
        The decorator can be used in the same context as "click_options()".
        This makes no sense, but works.
        Since the settings are passed as pos. args, the cli receives two
        instances in that case.
        """

        @click.command()
        @click_options(self.Settings, "test")
        @pass_settings
        def cli(s1: TestPassSettings.Settings, s2: TestPassSettings.Settings) -> None:
            assert s1 is s2

        invoke(cli, "--opt=spam")

    def test_pass_in_parent_context_argname(self, invoke: Invoke) -> None:
        """
        The decorator can be used in the same context as "click_options()".
        This makes no sense, but works.
        With an explicit argname, only one instance is passed.
        """

        @click.command()
        @click_options(self.Settings, "test", argname="le_settings")
        @pass_settings(argname="le_settings")
        def cli(*, le_settings: "TestPassSettings.Settings") -> None:
            assert le_settings == self.Settings("spam")

        invoke(cli, "--opt=spam")

    def test_combine_pass_settings_click_options(self, invoke: Invoke) -> None:
        """
        A sub command can receive the parent's options via "pass_settings"
        and define its own options at the same time.
        """

        @settings
        class SubSettings:
            sub: str = ""

        @click.group()
        @click_options(self.Settings, "test-main", argname="main")
        def cli(main: TestPassSettings.Settings) -> None:
            assert main == self.Settings("spam")

        @cli.command()
        @click_options(SubSettings, "test-sub", argname="sub")
        @pass_settings(argname="main")
        def cmd(main: TestPassSettings.Settings, sub: SubSettings) -> None:
            assert main == self.Settings("spam")
            assert sub == SubSettings("eggs")

        invoke(cli, "--opt=spam", "cmd", "--sub=eggs")


class TestClickConfig:
    """Tests for influencing the option declaration."""

    @pytest.mark.parametrize(
        "click_config",
        [None, {"param_decls": ("--opt/--no-opt",)}],
    )
    @pytest.mark.parametrize(
        "flag, value", [(None, True), ("--opt", True), ("--no-opt", False)]
    )
    def test_default_for_flag_has_on_and_off_switch(
        self,
        invoke: Invoke,
        click_config: Optional[dict],
        flag: Optional[str],
        value: bool,
    ) -> None:
        """
        The attrs default value is correctly used for flag options in all
        variants (no flag, on-flag, off-flag).
        """

        @settings
        class Settings:
            opt: bool = option(default=True, click=click_config)

        @click.command()
        @click_options(Settings, "test")
        def cli(settings: Settings) -> None:
            assert settings.opt is value

        if flag is None:
            result = invoke(cli)
        else:
            result = invoke(cli, flag)
        assert result.exit_code == 0

    @pytest.mark.parametrize(
        "flag, value", [(None, False), ("--opt", True), ("--no-opt", False)]
    )
    def test_create_a_flag_without_off_switch(
        self, invoke: Invoke, flag: Optional[str], value: bool
    ) -> None:
        """
        The "off"-flag for flag options can be removed.
        """
        click_config = {"param_decls": "--opt", "is_flag": True}

        @settings
        class Settings:
            opt: bool = option(default=False, click=click_config)

        @click.command()
        @click_options(Settings, "test")
        def cli(settings: Settings) -> None:
            assert settings.opt is value

        if flag is None:
            result = invoke(cli)
        else:
            result = invoke(cli, flag)

        if flag == "--no-opt":
            assert result.exit_code == 2
        else:
            assert result.exit_code == 0

    @pytest.mark.parametrize(
        "flag, value", [(None, False), ("-x", True), ("--exitfirst", True)]
    )
    def test_create_a_short_handle_for_a_flag(
        self, invoke: Invoke, flag: Optional[str], value: bool
    ) -> None:
        """
        Create a shorter handle for a command similar to pytest's -x.
        """
        click_config = {"param_decls": ("-x", "--exitfirst"), "is_flag": True}

        @settings
        class Settings:
            exitfirst: bool = option(default=False, click=click_config)

        @click.command()
        @click_options(Settings, "test")
        def cli(settings: Settings) -> None:
            assert settings.exitfirst is value

        if flag is None:
            result = invoke(cli)
        else:
            result = invoke(cli, flag)
        assert result.exit_code == 0

    @pytest.mark.parametrize("args, value", [([], False), (["--arg"], True)])
    def test_user_callback_is_executed(
        self, invoke: Invoke, args: List[str], value: bool
    ) -> None:
        """
        A user callback is only invoked if an argument was passed, but not if the
        default is used.
        """
        cb = mock.MagicMock(return_value=value)

        click_config = {"callback": cb}

        @settings
        class Settings:
            arg: bool = option(default=False, click=click_config)

        @click.command()
        @click_options(Settings, "test")
        def cli(settings: Settings) -> None:
            assert settings.arg is value

        result = invoke(cli, *args)
        assert result.exit_code == 0
        assert cb.call_count == int(bool(args))


class TestDecoratorFactory:
    """
    Tests for the decorator factory (e.g., for option groups).
    """

    @pytest.fixture
    def settings_cls(self) -> type:
        @settings
        class Nested1:
            """
            Docs for Nested1
            """  # noqa: D415

            a: int = 0

        @settings
        class Nested2:
            # Deliberately has no docstring!
            a: int = 0

        @settings
        class Settings:
            """
            Main docs
            """  # noqa: D415

            a: int = 0
            n1: Nested1 = Nested1()
            n2: Nested2 = Nested2()

        return Settings

    def test_click_option_factory(self, settings_cls: type, invoke: Invoke) -> None:
        """
        The ClickOptionFactory is the default.
        """

        @click.command()
        @click_options(settings_cls, "t")
        def cli1(settings: Any) -> None: ...

        @click.command()
        @click_options(
            settings_cls,
            "t",
            decorator_factory=cli_click.ClickOptionFactory(),
        )
        def cli2(settings: Any) -> None: ...

        r1 = invoke(cli1, "--help").output.splitlines()[1:]
        r2 = invoke(cli2, "--help").output.splitlines()[1:]
        assert r1 == r2

    def test_option_group_factory(self, settings_cls: type, invoke: Invoke) -> None:
        """
        Option groups can be created via the OptionGroupFactory.
        """

        @click.command()
        @click_options(
            settings_cls,
            "t",
            decorator_factory=cli_click.OptionGroupFactory(),
        )
        def cli(settings: Any) -> None: ...

        result = invoke(cli, "--help").output.splitlines()
        assert result == [
            "Usage: cli [OPTIONS]",
            "",
            "Options:",
            "  Main docs: ",
            "    --a INTEGER       [default: 0]",
            "  Docs for Nested1: ",
            "    --n1-a INTEGER    [default: 0]",
            "  Nested2 options: ",
            "    --n2-a INTEGER    [default: 0]",
            "  --help              Show this message and exit.",
        ]

    def test_not_installed(self, unimport: Callable[[str], None]) -> None:
        """
        The factory checks if click-option-group is installed.
        """
        unimport("click_option_group")
        with pytest.raises(ModuleNotFoundError):
            cli_click.OptionGroupFactory()


@pytest.mark.parametrize(
    "factory",
    [None, cli_click.ClickOptionFactory(), cli_click.OptionGroupFactory()],
)
def test_show_envvar_in_help(
    factory: Optional[cli_click.DecoratorFactory], invoke: Invoke
) -> None:
    """
    An option's help can optionally show the env var that will be loaded.
    """

    @settings
    class Settings:
        a: str = "spam"
        b: str = secret(default="eggs")

    @click.command()
    @click_options(
        Settings,
        default_loaders("test"),
        decorator_factory=factory,
        show_envvars_in_help=True,
    )
    def cli(settings: Settings) -> None: ...

    result = invoke(cli, "--help")
    if isinstance(factory, cli_click.OptionGroupFactory):
        print(result.output)
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  Settings options: \n"
            "    --a TEXT          [env var: TEST_A; default: spam]\n"
            "    --b TEXT          [env var: TEST_B; default: (*******)]\n"
            "  --help              Show this message and exit.\n"
        )
    else:
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --a TEXT  [env var: TEST_A; default: spam]\n"
            "  --b TEXT  [env var: TEST_B; default: (*******)]\n"
            "  --help    Show this message and exit.\n"
        )
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "factory",
    [None, cli_click.ClickOptionFactory(), cli_click.OptionGroupFactory()],
)
def test_click_no_load_envvar(
    factory: Optional[cli_click.DecoratorFactory],
    invoke: Invoke,
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    The "show_envvars_in_help" option does not cause Click to load settings
    from envvars.
    """
    tmp_path.joinpath("settings.toml").write_text('[test]\na = "x"\n')
    spath = tmp_path.joinpath("settings2.toml")
    spath.write_text('[test]\na = "spam"\n')
    monkeypatch.setenv("TEST_A", "onoes")

    @settings
    class Settings:
        a: str = "spam"

    # Reverse loaders so that env loader is used first and the file loader
    # is used last (and thus has priority)
    loaders = list(reversed(default_loaders("test", [spath])))

    @click.command()
    @click_options(
        Settings,
        loaders,
        decorator_factory=factory,
        show_envvars_in_help=True,
    )
    def cli(settings: Settings) -> None:
        print(settings.a)

    result = invoke(cli)
    # If click read from the envvar, the output would be "onoes"
    assert result.output == "spam\n"
    assert result.exit_code == 0


def test_resolve_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, invoke: Invoke
) -> None:
    """
    Relative paths passed via the command line are resolved based on the user's CWD.
    """

    @settings
    class Settings:
        a: Path = Path("default")
        b: Path = Path("default")  # Load from file
        c: Path = Path("default")  # Load from env var
        d: Path = Path("default")  # Load from cli arg

    spath = tmp_path.joinpath("settings.toml")
    spath.write_text('[test]\nb = "file"\n')
    monkeypatch.setenv("TEST_C", "env")

    # chdir *before* creating the CLI, b/c it will load the defaults immediately:
    subdir = tmp_path.joinpath("sub")
    subdir.mkdir()
    monkeypatch.chdir(subdir)

    result = Settings()  # Will be update by the CLI

    @click.command()
    @click_options(Settings, default_loaders("test", [spath]))
    def cli(settings: Settings) -> None:
        nonlocal result
        result = settings

    invoke(cli, "--d", "arg")
    assert result == Settings(
        a=subdir.joinpath("default"),
        b=spath.parent.joinpath("file"),
        c=subdir.joinpath("env"),
        d=subdir.joinpath("arg"),
    )


def test_multiple_invocations(invoke: Invoke) -> None:
    """
    A CLI function can be invoked multiple times w/o carrying state from call to call.
    """

    @settings
    class Settings:
        o: int = 0

    loaded_settings: List[Settings] = []

    @click.command()
    @click_options(Settings, "example")
    def cli(settings: Settings) -> None:
        loaded_settings.append(settings)

    # The order of these invocations is important:
    invoke(cli, "--o=3")
    invoke(cli)
    assert loaded_settings == [Settings(3), Settings(0)]


def test_pydantic_secrets(invoke: Invoke) -> None:
    """
    Tests for pydantic secrets handling together with click.
    """

    class Settings(pydantic.BaseModel):
        secret: pydantic.SecretStr = pydantic.Field(
            default=pydantic.SecretStr("secret-default"),
        )

    default_settings = Settings()

    assert default_settings.secret.get_secret_value() == "secret-default"

    loaded_settings: List[Settings] = []

    @click.command()
    @click_options(Settings, "example")
    def cli(settings: Settings) -> None:
        loaded_settings.append(settings)

    invoke(cli)
    invoke(cli, "--secret=secret-string")

    assert loaded_settings == [
        Settings(),
        Settings(secret=pydantic.SecretStr("secret-string")),
    ]
    assert loaded_settings[0].secret.get_secret_value() == "secret-default"
    assert loaded_settings[1].secret.get_secret_value() == "secret-string"


def test_lazy_load_defaults(invoke: Invoke, monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Defaults for CLIs can be loaded lazily (when the CLI is invoked, not when it is
    constructed).
    """

    @settings
    class Settings:
        o1: int
        o2: int = 0

    loaded_settings: List[Settings] = []

    monkeypatch.setenv("EXAMPLE_O2", "1")

    @click.command()
    @click_options(Settings, "example", reload_settings_on_invoke=True)
    def cli(settings: Settings) -> None:
        loaded_settings.append(settings)

    # The order of these invocations is important:
    result = invoke(cli, "--help")
    assert result.output == (
        "Usage: cli [OPTIONS]\n"
        "\n"
        "Options:\n"
        "  --o1 INTEGER  [required]\n"
        "  --o2 INTEGER  [default: 1]\n"
        "  --help        Show this message and exit.\n"
    )
    invoke(cli, "--o1=0")

    # Setting the env var to another value would have no effect w/o lazy loading:
    monkeypatch.setenv("EXAMPLE_O2", "2")
    invoke(cli, "--o1=0")

    assert loaded_settings == [Settings(0, 1), Settings(0, 2)]



"""
Tests for "typed_settings.processors".
"""

from typing import Callable, Type, Union

import jinja2
import pytest

from typed_settings import processors, settings
from typed_settings.cls_utils import deep_options


@settings
class ChildA:
    """A nested class."""

    aa: str
    ab: str


@settings
class ChildB:
    """A nested class."""

    ba: str
    bb: str


@settings
class Settings:
    """A parent class."""

    a: ChildA
    b: ChildB
    c: str
    d: str


OPTIONS = deep_options(Settings)


@pytest.mark.parametrize(
    "value, expected",
    [
        ("spam", "spam"),
        ("raw://spam", "spam"),
        ("helper://echo spam", "spam"),
        ("script://echo spam", "spam"),
        # Only process a value once:
        ("raw://script://echo spam", "script://echo spam"),
        ("script://echo 'raw://spam'", "raw://spam"),
    ],
)
def test_url_processor(value: str, expected: str) -> None:
    """
    The processor handles the configured protocols and changes the dict
    in-place.
    """

    @settings
    class Settings:
        o: str

    uh = processors.UrlProcessor(
        {
            "raw://": processors.handle_raw,
            "helper://": processors.handle_script,
            "script://": processors.handle_script,
        }
    )
    result = uh({"o": value}, Settings, deep_options(Settings))
    assert result == {"o": expected}


def test_handle_raw() -> None:
    """
    The handler returns the value unchanged.
    """
    result = processors.handle_raw("spam", "raw://")
    assert result == "spam"


def test_handle_script() -> None:
    """
    The script is run and its output returned stripped.
    """
    result = processors.handle_script("echo spam", "script://")
    assert result == "spam"


@pytest.mark.parametrize(
    "cmd, code, stdout, stderr",
    [
        ("xyz", 127, "", ".*/bin/sh.*xyz.*not found\n"),
        ("echo a; echo b 1>&2; exit 1", 1, "a\n", "b\n"),
    ],
)
def test_handle_script_error(cmd: str, code: int, stdout: str, stderr: str) -> None:
    """
    Raise ValueError if the command cannot be found or fails.  Include stdout
    and stderr in exception.
    """
    msg = (
        f"Helper script failed: script://{cmd}\n"
        f"EXIT CODE: {code}\n"
        f"STDOUT:\n{stdout}"
        f"STDERR:\n{stderr}"
    )
    with pytest.raises(ValueError, match=msg):
        processors.handle_script(cmd, "script://")


def test_handle_op(mock_op: None) -> None:
    """
    The 1Password handler retrievs the secret from the "op" CLI.
    """
    result = processors.handle_op("Test/Test/password", "op://")
    assert result == "eggs"


class TestFormatProcessor:
    """
    Tests for "format_processor()".
    """

    DATA = [
        pytest.param(
            {"c": "spam"},
            {"c": "spam"},
            id="str-no-template",
        ),
        pytest.param(
            {"c": 3},
            {"c": 3},
            id="int-no-template",
        ),
        pytest.param(
            {"c": "spam", "d": "{c}"},
            {"c": "spam", "d": "spam"},
            id="template-after-value",
        ),
        pytest.param(
            {"c": "{d}", "d": "spam"},
            {"c": "spam", "d": "spam"},
            id="template-before-value",
        ),
        pytest.param(
            {"a": {"aa": "spam"}, "c": "{a[aa]}"},
            {"a": {"aa": "spam"}, "c": "spam"},
            id="refer-to-child",
        ),
        pytest.param(
            {"a": {"aa": "{c}"}, "c": "spam"},
            {"a": {"aa": "spam"}, "c": "spam"},
            id="refer-to-parent",
        ),
        pytest.param(
            {"a": {"aa": "{a[ab]}", "ab": "spam"}},
            {"a": {"aa": "spam", "ab": "spam"}},
            id="refer-to-self",
        ),
        pytest.param(
            {"a": {"aa": "{b[ba]}"}, "b": {"ba": "spam"}},
            {"a": {"aa": "spam"}, "b": {"ba": "spam"}},
            id="refer-to-sibling",
        ),
        pytest.param(
            {
                "a": {
                    "aa": "{c}",
                    "ab": "{b[ba]}",
                },
                "b": {
                    "ba": "{a[aa]}",
                    "bb": "{b[ba]}",
                },
                "c": "{d}",
                "d": "spam",
            },
            {
                "a": {
                    "aa": "spam",
                    "ab": "spam",
                },
                "b": {
                    "ba": "spam",
                    "bb": "spam",
                },
                "c": "spam",
                "d": "spam",
            },
            id="chain-templates",
        ),
        pytest.param(
            {"c": "{d}"},
            {"c": "{d}"},
            id="simple-value-not-defined",
        ),
        pytest.param(
            {"c": "{a.aa}", "a": {"aa": {"spam"}}},
            {"c": "{a.aa}", "a": {"aa": {"spam"}}},
            id="invalid-attr-access",
        ),
        pytest.param(
            {"c": "spam", "d": "{c"},
            {"c": "spam", "d": "{c"},
            id="invalid-template-1",
        ),
        pytest.param(
            {"c": "spam", "d": "{{c}"},
            {"c": "spam", "d": "{{c}"},
            id="invalid-template-2",
        ),
        pytest.param(
            {"c": "spam", "d": "{{}{}}"},
            {"c": "spam", "d": "{{}{}}"},
            id="invalid-template-3",
        ),
        pytest.param(
            {"c": "spam", "d": "{{}}"},
            {"c": "spam", "d": "{}"},
            id="invalid-template-4",
        ),
        pytest.param(
            {"c": "{d}", "d": "{c}"},
            RecursionError,
            id="recursion",
        ),
        pytest.param(
            {"c": "{c}"},
            {"c": "{c}"},
            id="self-recursion",
        ),
    ]

    @pytest.mark.parametrize("data, expected", DATA)
    def test_format_processor(
        self, data: dict, expected: Union[dict, Type[Exception]]
    ) -> None:
        """
        Values are recursively rendered with "str.format()" and errors are
        never raised.
        """
        if isinstance(expected, dict):
            result = processors.FormatProcessor()(data, Settings, OPTIONS)
            assert result == expected
        else:
            with pytest.raises(expected):
                processors.FormatProcessor()(data, Settings, OPTIONS)


class TestJinjaProcessor:
    """
    Tests for "jinja_processor()".
    """

    DATA = [
        pytest.param(
            {"c": "spam"},
            {"c": "spam"},
            id="str-no-template",
        ),
        pytest.param(
            {"c": 3},
            {"c": 3},
            id="int-no-template",
        ),
        pytest.param(
            {"c": "spam", "d": "{{ c }}"},
            {"c": "spam", "d": "spam"},
            id="template-after-value",
        ),
        pytest.param(
            {"c": "{{ d }}", "d": "spam"},
            {"c": "spam", "d": "spam"},
            id="simple-template-before-value",
        ),
        pytest.param(
            {"a": {"aa": "spam"}, "c": "{{ a.aa }}"},
            {"a": {"aa": "spam"}, "c": "spam"},
            id="refer-to-child",
        ),
        pytest.param(
            {"a": {"aa": "spam"}, "c": "{{ a['aa'] }}"},
            {"a": {"aa": "spam"}, "c": "spam"},
            id="refer-to-child-item-access",
        ),
        pytest.param(
            {"a": {"aa": "{{ c }}"}, "c": "spam"},
            {"a": {"aa": "spam"}, "c": "spam"},
            id="refer-to-parent",
        ),
        pytest.param(
            {"a": {"aa": "{{ a.ab }}", "ab": "spam"}},
            {"a": {"aa": "spam", "ab": "spam"}},
            id="refer-to-self",
        ),
        pytest.param(
            {"a": {"aa": "{{ b.ba }}"}, "b": {"ba": "spam"}},
            {"a": {"aa": "spam"}, "b": {"ba": "spam"}},
            id="refer-to-sibling",
        ),
        pytest.param(
            {
                "a": {
                    "aa": "{{ c }}",
                    "ab": "{{ b.ba }}",
                },
                "b": {
                    "ba": "{{ a.aa }}",
                    "bb": "{{ b.ba }}",
                },
                "c": "{{ d }}",
                "d": "spam",
            },
            {
                "a": {
                    "aa": "spam",
                    "ab": "spam",
                },
                "b": {
                    "ba": "spam",
                    "bb": "spam",
                },
                "c": "spam",
                "d": "spam",
            },
            id="chain-templates",
        ),
        pytest.param(
            {
                "a": {
                    "aa": "{{ c }}",
                    "ab": "{{ b.ba }}",
                },
                "b": {
                    "ba": "{% if a.aa == 'spam' %}x{% else %}y{% endif %}",
                    "bb": "{{ b.ba }}",
                },
                "c": "{{ d }}",
                "d": "spam",
            },
            {
                "a": {
                    "aa": "spam",
                    "ab": "x",
                },
                "b": {
                    "ba": "x",
                    "bb": "x",
                },
                "c": "spam",
                "d": "spam",
            },
            id="if-expressions",
        ),
        pytest.param(
            {"c": "{}"},
            {"c": "{}"},
            id="format-string",
        ),
        pytest.param(
            {"c": "{{}}"},
            {"c": "{{}}"},
            id="empty-template",
        ),
        pytest.param(
            {"c": "{{ d }}"},
            {"c": ""},
            id="value-not-defined",
        ),
        pytest.param(
            {"c": "spam", "d": "{{ c"},
            {"c": "spam", "d": "{{ c"},
            id="invalid-template-1",
        ),
        pytest.param(
            {"c": "spam", "d": "{{c}"},
            {"c": "spam", "d": "{{c}"},
            id="invalid-template-2",
        ),
        pytest.param(
            {"c": "spam", "d": "{{}{}}"},
            {"c": "spam", "d": "{{}{}}"},
            id="invalid-template-3",
        ),
        pytest.param(
            {"c": "{{ d }}", "d": "{{ c }}"},
            RecursionError,
            id="recursion",
        ),
        pytest.param(
            {"c": "{{ c }}"},
            RecursionError,
            id="self-recursion",
        ),
    ]

    DATA_WITH_CUSTOM_JINJA = [
        *DATA,
        pytest.param(
            {"d": "{{ foo('bar') }}"}, {"d": "foo_bar"}, id="custom-jinja-function"
        ),
        pytest.param(
            {"d": "{{ 'foo' | bar(2) }}"}, {"d": "foofoo"}, id="custom-jinja-filter"
        ),
    ]

    @pytest.mark.parametrize("data, expected", DATA)
    def test_jinja_processor(
        self, data: dict, expected: Union[dict, Type[Exception]]
    ) -> None:
        """
        Values are recursively rendered with Jinja2 and an error is never
        raised.
        """
        if isinstance(expected, dict):
            result = processors.JinjaProcessor()(data, Settings, OPTIONS)
            assert result == expected
        else:
            with pytest.raises(expected):
                processors.JinjaProcessor()(data, Settings, OPTIONS)

    @pytest.mark.parametrize("data, expected", DATA_WITH_CUSTOM_JINJA)
    def test_jinja_processor_custom_environment(
        self, data: dict, expected: Union[dict, Type[Exception]]
    ):
        """
        Custom jinja environments with user extensions can be passed to the
        JinjaProcessor.
        """
        env = jinja2.Environment()  # noqa: S701
        env.globals["foo"] = lambda x: f"foo_{x}"
        env.filters["bar"] = lambda val, times: val * times

        if isinstance(expected, dict):
            result = processors.JinjaProcessor(environment=env)(data, Settings, OPTIONS)
            assert result == expected
        else:
            with pytest.raises(expected):
                processors.JinjaProcessor(environment=env)(data, Settings, OPTIONS)

    def test_jinja_not_installed(self, unimport: Callable[[str], None]) -> None:
        unimport("jinja2")
        with pytest.raises(ModuleNotFoundError, match="not installed"):
            processors.JinjaProcessor()



"""
Tests for "typed_settings.dict_utils".
"""

import dataclasses
from typing import Any, Dict, List, Type, Union

import attrs
import pytest

from typed_settings import dict_utils, types


def test_iter_settings() -> None:
    """
    "iter_settings()" iterates the settings.  It ignores invalid settings keys
    or non-existing settings.
    """

    @dataclasses.dataclass
    class ListSettings:
        w: int

    option_infos = [
        types.OptionInfo(
            parent_cls=type,
            path=path,
            cls=int,
            default=attrs.NOTHING,
            has_no_default=True,
            default_is_factory=False,
        )
        for path in ["a", "b.x", "b.y", "c"]
    ]
    option_infos.append(
        types.OptionInfo(
            parent_cls=type,
            path="u",
            cls=List[int],
            default=attrs.NOTHING,
            has_no_default=True,
            default_is_factory=False,
        )
    )
    option_infos.append(
        types.OptionInfo(
            parent_cls=type,
            path="v",
            cls=List[ListSettings],
            default=attrs.NOTHING,
            has_no_default=True,
            default_is_factory=False,
        )
    )
    option_infos.append(
        types.OptionInfo(
            parent_cls=type,
            path="x",
            cls=List[str],
            default=attrs.NOTHING,
            has_no_default=True,
            default_is_factory=False,
        )
    )
    settings = {
        "a": 0,
        "b": {
            "y": 1,
        },
        "z": 2,
        "u": [4, 5],
        "v": [
            {
                "w": 6,
            }
        ],
        "x": "hello:world",
    }
    result = list(dict_utils.iter_settings(settings, tuple(option_infos)))
    assert result == [
        ("a", 0),
        ("b.y", 1),
        ("u.0", 4),
        ("u.1", 5),
        ("v.0.w", 6),
        ("x", "hello:world"),
    ]


@pytest.mark.parametrize(
    "path, expected",
    [
        ("a", 1),
        ("b.c", 2),
        ("b.d.e", 3),
        ("x", KeyError),
        ("b.x", KeyError),
        ("u.0", 4),
        ("u.1", 5),
        ("u.2", IndexError),
    ],
)
def test_get_path(path: str, expected: Union[int, Type[Exception]]) -> None:
    """Tests for get_path()."""
    dct = {
        "a": 1,
        "b": {
            "c": 2,
            "d": {
                "e": 3,
            },
        },
        "u": [4, 5],
    }
    if isinstance(expected, int):
        assert dict_utils.get_path(dct, path) == expected
    else:
        pytest.raises(expected, dict_utils.get_path, dct, path)


def test_set_path() -> None:
    """We can set arbitrary paths, nested dicts will be created as needed."""
    dct: Dict[str, Any] = {}
    dict_utils.set_path(dct, "a", 0)
    dict_utils.set_path(dct, "a", 1)
    dict_utils.set_path(dct, "b.d.e", 3)
    dict_utils.set_path(dct, "b.c", 2)
    dict_utils.set_path(dct, "u.0", 4)
    dict_utils.set_path(dct, "u.1", 5)
    dict_utils.set_path(dct, "u", [None, None])
    dict_utils.set_path(dct, "u.1", 5)
    dict_utils.set_path(dct, "v", [{}])
    dict_utils.set_path(dct, "v.0.w", 6)
    assert dct == {
        "a": 1,
        "b": {
            "c": 2,
            "d": {
                "e": 3,
            },
        },
        "u": [None, 5],
        "v": [{"w": 6}],
    }


def test_merge_settings() -> None:
    """
    When settings are merged, merging only applies to keys for options, not list or
    dict values.
    """
    option_infos = tuple(
        types.OptionInfo(
            parent_cls=type,
            path=path,
            cls=tuple,
            default=attrs.NOTHING,
            has_no_default=True,
            default_is_factory=False,
        )
        for path in ["1a", "1b.2a", "1b.2b.3a", "1b.2b.3b", "1c", "1d", "1e"]
    )
    d1 = types.LoadedSettings(
        {
            "1a": 3,
            "1b": {"2a": "spam", "2b": {"3a": "foo"}},
            "1c": [{"2a": 3.14}, {"2b": 34.3}],  # Do not merge lists
            "1d": 4,
            "1e": {"default": "default"},  # Do not merge dicts
        },
        types.LoaderMeta("l1"),
    )
    d2 = types.LoadedSettings(
        {
            "1b": {"2a": "eggs", "2b": {"3b": "bar"}},
            "1c": [{"2a": 23}, {"2b": 34.3}],
            "1d": 5,
            "1e": {"update": "value"},
        },
        types.LoaderMeta("l2"),
    )
    result = dict_utils.merge_settings(option_infos, [d1, d2])
    assert result == {
        "1a": types.LoadedValue(3, types.LoaderMeta("l1")),
        "1b.2a": types.LoadedValue("eggs", types.LoaderMeta("l2")),
        "1b.2b.3a": types.LoadedValue("foo", types.LoaderMeta("l1")),
        "1b.2b.3b": types.LoadedValue("bar", types.LoaderMeta("l2")),
        "1c": types.LoadedValue([{"2a": 23}, {"2b": 34.3}], types.LoaderMeta("l2")),
        "1d": types.LoadedValue(5, types.LoaderMeta("l2")),
        "1e": types.LoadedValue({"update": "value"}, types.LoaderMeta("l2")),
    }


def test_update_settings() -> None:
    """
    When updating settings, the input remains unmodified and an updated dopy is
    returned.
    """
    merged = {
        "1a": types.LoadedValue(1, types.LoaderMeta("l1")),
        "1b": types.LoadedValue(1, types.LoaderMeta("l1")),
        "1c": types.LoadedValue(1, types.LoaderMeta("l1")),
    }
    result = dict_utils.update_settings(merged, {"1b": 2})
    assert merged == {
        "1a": types.LoadedValue(1, types.LoaderMeta("l1")),
        "1b": types.LoadedValue(1, types.LoaderMeta("l1")),
        "1c": types.LoadedValue(1, types.LoaderMeta("l1")),
    }
    assert result == {
        "1a": types.LoadedValue(1, types.LoaderMeta("l1")),
        "1b": types.LoadedValue(2, types.LoaderMeta("l1")),
        "1c": types.LoadedValue(1, types.LoaderMeta("l1")),
    }


def test_flat2nested() -> None:
    """
    "flat2nested" converts a flat dict "option_path: value" to a nested dict.  The
    keys no longer contain dots.
    """
    merged = {
        "1a": types.LoadedValue(3, types.LoaderMeta("l1")),
        "1b.2a": types.LoadedValue("eggs", types.LoaderMeta("l2")),
        "1b.2b.3a": types.LoadedValue("foo", types.LoaderMeta("l1")),
        "1b.2b.3b": types.LoadedValue("bar", types.LoaderMeta("l2")),
        "1c": types.LoadedValue([{"2a": 23}, {"2b": 34.3}], types.LoaderMeta("l2")),
        "1d": types.LoadedValue(5, types.LoaderMeta("l2")),
        "1e": types.LoadedValue({"update": "value"}, types.LoaderMeta("l2")),
    }
    result = dict_utils.flat2nested(merged)
    assert result == {
        "1a": 3,
        "1b": {"2a": "eggs", "2b": {"3a": "foo", "3b": "bar"}},
        "1c": [{"2a": 23}, {"2b": 34.3}],
        "1d": 5,
        "1e": {"update": "value"},
    }



"""
Tests for "typed_settings._core".
"""

import json
import logging
import re
from pathlib import Path
from typing import Any, Dict, List, cast

import attrs
import cattrs
import pytest

from typed_settings import _core, dict_utils, exceptions
from typed_settings._compat import PY_310
from typed_settings.cls_attrs import option, settings
from typed_settings.converters import (
    default_converter,
    get_default_cattrs_converter,
    register_strlist_hook,
)
from typed_settings.loaders import DictLoader, EnvLoader, FileLoader, Loader, TomlFormat
from typed_settings.types import (
    LoadedSettings,
    LoadedValue,
    LoaderMeta,
    OptionList,
    SettingsClass,
    SettingsDict,
)


@settings(frozen=True)
class Host:
    """Host settings."""

    name: str
    port: int


@settings(frozen=True)
class Settings:
    """Main settings."""

    host: "Host"  # Assert that types are resolved
    url: str
    default: int = option(default=3, validator=attrs.validators.gt(0))


class TestLoadSettings:
    """Tests for load_settings()."""

    config = """[example]
        url = "https://example.com"
        [example.host]
        name = "example.com"
        port = 443
    """

    @pytest.fixture
    def config_files(self, tmp_path: Path) -> List[Path]:
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(self.config)
        return [config_file]

    @pytest.fixture
    def env_vars(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("EXAMPLE_HOST_PORT", "42")

    @pytest.fixture
    def loaders(self, config_files: List[Path], env_vars: None) -> List[Loader]:
        return [
            FileLoader(
                formats={"*.toml": TomlFormat("example")},
                files=config_files,
            ),
            EnvLoader(prefix="EXAMPLE_"),
        ]

    def test__load_settings(self, loaders: List[Loader], tmp_path: Path) -> None:
        """
        "_load_settings()" is the internal core loader and takes a list of
        options instead of a normal settings class.  It returns a dict and
        not a settings instance.
        """
        cwd = Path.cwd()
        state = _core.SettingsState(Settings, loaders, [], default_converter(), cwd)
        settings = _core._load_settings(state)
        assert settings == {
            "host.name": LoadedValue(
                "example.com",
                LoaderMeta(
                    f"FileLoader[{tmp_path.joinpath('settings.toml')}]",
                    base_dir=tmp_path,
                ),
            ),
            "host.port": LoadedValue("42", LoaderMeta("EnvLoader", base_dir=cwd)),
            "url": LoadedValue(
                "https://example.com",
                LoaderMeta(
                    f"FileLoader[{tmp_path.joinpath('settings.toml')}]",
                    base_dir=tmp_path,
                ),
            ),
            "default": LoadedValue(3, LoaderMeta("_DefaultsLoader", base_dir=cwd)),
        }

    def test_load_settings(self, loaders: List[Loader]) -> None:
        """
        The "load_settings()" works like "_load_settings" but takes a settings
        class and returns an instance of it.
        """
        settings = _core.load_settings(
            cls=Settings,
            loaders=loaders,
        )
        assert settings == Settings(
            url="https://example.com",
            default=3,
            host=Host(
                name="example.com",
                port=42,
            ),
        )

    def test_load(self, config_files: List[Path], env_vars: None) -> None:
        """
        The "load()" shortcut automaticaly defines a file loader and an
        env loader.  Section and env var names are derived from the app name.
        """
        settings = _core.load(
            cls=Settings,
            appname="example",
            config_files=config_files,
        )
        assert settings == Settings(
            url="https://example.com",
            default=3,
            host=Host(
                name="example.com",
                port=42,  # Loaded from env var
            ),
        )

    def test_explicit_section(self, tmp_path: Path) -> None:
        """
        The automatically derived config section name name can be overriden.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[le-section]
            spam = "eggs"
        """
        )

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings,
            appname="example",
            config_files=[config_file],
            config_file_section="le-section",
        )
        assert result == Settings(spam="eggs")

    def test_explicit_files_var(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        The automatically derived settings files var name can be overriden.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[example]
            spam = "eggs"
        """
        )

        monkeypatch.setenv("LE_SETTINGS", str(config_file))

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings,
            appname="example",
            config_files=[],
            config_files_var="LE_SETTINGS",
        )
        assert result == Settings(spam="eggs")

    def test_no_files_var(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Setting config files via an env var can be disabled.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[example]
            spam = "eggs"
        """
        )

        monkeypatch.setenv("EXAMPLE_SETTINGS", str(config_file))

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings,
            appname="example",
            config_files=[],
            config_files_var=None,
        )
        assert result == Settings(spam="")

    def test_env_var_dash_underscore(
        self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path
    ) -> None:
        """
        Dashes in the appname get replaced with underscores for the settings
        fiels var name.
        """

        @settings(frozen=True)
        class Settings:
            option: bool = True

        sf = tmp_path.joinpath("settings.toml")
        sf.write_text("[a-b]\noption = false\n")
        monkeypatch.setenv("A_B_SETTINGS", str(sf))

        result = _core.load(Settings, appname="a-b")
        assert result == Settings(option=False)

    def test_implicit_env_prefix_appname_dash(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Dashes in the appname get replaced for the env vars.
        """

        @settings(frozen=True)
        class Settings:
            option: bool = False

        monkeypatch.setenv("A_B_OPTION", "1")

        result = _core.load(Settings, appname="a-b")
        assert result == Settings(option=True)

    def test_explicit_env_prefix(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("P_SPAM", "spam")

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings, appname="example", config_files=[], env_prefix="P_"
        )
        assert result == Settings(spam="spam")

    def test_disable_env_vars(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("EXAMPLE_SPAM", "spam")

        @settings(frozen=True)
        class Settings:
            spam: str = ""

        result = _core.load(
            cls=Settings, appname="example", config_files=[], env_prefix=None
        )
        assert result == Settings(spam="")

    def test_load_nested_settings_by_default(self) -> None:
        """
        Instantiate nested settings with default settings and pass it to the
        parent settings even if no nested settings are defined in a config
        file or env var.

        Otherwise, the parent classed needed to set a default_factory for
        creating a nested settings instance.
        """

        @settings
        class Nested:
            a: int = 3
            b: str = "spam"

        @settings
        class Settings:
            nested: Nested

        s = _core.load(Settings, "test")
        assert s == Settings(Nested())

    def test_default_factories(self) -> None:
        """
        The default value "attr.Factory" is handle as if "attr.NOTHING" was
        set.

        See: https://gitlab.com/sscherfke/typed-settings/-/issues/6
        """

        @settings
        class S:
            opt: List[int] = option(factory=list)

        result = _core.load(S, "t")
        assert result == S()

    def test_custom_converter(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """
        A custom cattr converter can be used in "load_settings()".
        """

        class Test:
            def __init__(self, x: int):
                self.attr = x

            def __eq__(self, other: object) -> bool:
                return self.attr == other.attr  # type: ignore

        @settings
        class Settings:
            opt: Test

        converter = cattrs.Converter()
        converter.register_structure_hook(
            Test, lambda v, t: v if isinstance(v, Test) else Test(int(v))
        )

        result = _core.load_settings(
            Settings, [DictLoader({"opt": "42"})], converter=converter
        )
        assert result == Settings(Test(42))

    def test_custom_option_converter(self) -> None:
        """
        Converters defined for individual options are being used.
        """

        @settings
        class Settings:
            opt: str = option(converter=lambda v: f"converted:{v}")

        result = _core.load_settings(Settings, [DictLoader({"opt": "a"})])
        assert result == Settings("converted:a")

    @pytest.mark.parametrize(
        "vals, kwargs",
        [
            (["3,4,42", "spam,eggs"], {"sep": ","}),
            (["3:4:42", "spam:eggs"], {"sep": ":"}),
            (['[3,4,"42"]', '["spam","eggs"]'], {"fn": json.loads}),
        ],
    )
    def test_load_list_from_env(
        self,
        vals: List[str],
        kwargs: Dict[str, Any],
        loaders: List[Loader],
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        Lists can be loaded from env vars.
        """
        c = get_default_cattrs_converter()
        register_strlist_hook(c, **kwargs)

        @settings
        class Settings:
            x: List[int]
            y: List[Path]
            z: List[int]

        # The str2list hook should not interfere with loading of "normal"
        # lists.
        sf = tmp_path.joinpath("settings.toml")
        sf.write_text("[example]\nz = [1, 2]\n")

        monkeypatch.setenv("EXAMPLE_X", vals[0])
        monkeypatch.setenv("EXAMPLE_Y", vals[1])

        result = _core.load_settings(Settings, loaders, converter=c)
        cwd = Path.cwd().joinpath
        assert result == Settings(x=[3, 4, 42], y=[cwd("spam"), cwd("eggs")], z=[1, 2])

    def test_load_empty_cls(self) -> None:
        """
        Empty classes are no special case.
        """

        @settings
        class Settings:
            pass

        result = _core.load(Settings, "example")
        assert result == Settings()

    def test_processors_applied(self, loaders: List[Loader]) -> None:
        """
        Processors are applied to the loaded settings (including the defaults).
        """

        def p1(
            settings_dict: SettingsDict,
            settings_cls: SettingsClass,
            options: OptionList,
        ) -> SettingsDict:
            assert settings_dict == {
                "url": "https://example.com",
                "default": 3,
                "host": {
                    "name": "example.com",
                    "port": "42",
                },
            }
            settings_dict["url"] = "spam"
            return settings_dict

        def p2(
            settings_dict: SettingsDict,
            settings_cls: SettingsClass,
            options: OptionList,
        ) -> SettingsDict:
            assert settings_dict["url"] == "spam"
            settings_dict["host"]["port"] = "2"
            return settings_dict

        settings = _core.load_settings(
            cls=Settings,
            loaders=loaders,
            processors=[p1, p2],
        )
        assert settings == Settings(
            url="spam",
            default=3,
            host=Host(
                name="example.com",
                port=2,
            ),
        )

    def test_resolve_paths(
        self, loaders: List[Loader], tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Field converters are preferred over the cattrs converter.
        They can be used to resolves paths relative to the loader's cwd.

        See: https://gitlab.com/sscherfke/typed-settings/-/issues/30
        """

        @settings
        class S:
            a: Path
            b: Path
            c: Path
            d: Path

        d1 = tmp_path.joinpath("d1")
        d1.mkdir()
        c1 = d1.joinpath("s.toml")
        c1.write_text('[example]\na = "f0"\nb = "f1"\n')
        d2 = tmp_path.joinpath("d2")
        d2.mkdir()
        c2 = d2.joinpath("s.toml")
        c2.write_text('[example]\nc = "f2"\n')
        monkeypatch.setenv("EXAMPLE_D", "f3")
        cast(FileLoader, loaders[0]).files = [c1, c2]

        result = _core.load_settings(cls=S, loaders=loaders)
        assert result == S(
            a=d1.joinpath("f0"),
            b=d1.joinpath("f1"),
            c=d2.joinpath("f2"),
            d=Path.cwd().joinpath("f3"),
        )

    @pytest.mark.skipif(
        not PY_310, reason="Error messages differ a bit on older versions"
    )
    @pytest.mark.parametrize(
        "settings, err",
        [
            (
                {"url": "u"},
                (
                    "3 errors occured while converting the loaded option values to an "
                    "instance of 'Settings':\n"
                    "- No value set for required option 'host.name'\n"
                    "- No value set for required option 'host.port'\n"
                    "- Could not convert loaded settings: "
                    'TypeError("Settings.__init__() missing 1 required positional '
                    "argument: 'host'\")"
                ),
            ),
            (
                {"host": {"name": "h"}, "url": "u"},
                (
                    "2 errors occured while converting the loaded option values to an "
                    "instance of 'Settings':\n"
                    "- No value set for required option 'host.port'\n"
                    '- Could not convert loaded settings: TypeError("Host.__init__() '
                    "missing 1 required positional argument: 'port'\")"
                ),
            ),
            (
                {"host": {"name": "h", "port": "spam"}, "url": "u"},
                (
                    "3 errors occured while converting the loaded option values to an "
                    "instance of 'Settings':\n"
                    "- Could not convert value 'spam' for option 'host.port' from "
                    'loader test: ValueError("invalid literal for int() with base 10: '
                    "'spam'\")\n"
                    "- No value set for required option 'host.port'\n"
                    "- Could not convert loaded settings: "
                    'TypeError("Host.__init__() missing 1 required positional '
                    "argument: 'port'\")"
                ),
            ),
            (
                {"host": {"name": "h", "port": 1}, "url": "u", "default": -1},
                (
                    "1 errors occured while converting the loaded option values to an "
                    "instance of 'Settings':\n"
                    "- Could not convert loaded settings: "
                    "ValueError(\"'default' must be > 0: -1\")"
                ),
            ),
        ],
    )
    def test_convert_errors(self, settings: dict, err: str) -> None:
        state = _core.SettingsState(Settings, [], [], default_converter(), Path())
        meta = LoaderMeta("test")
        merged = dict_utils.merge_settings(
            state.options, [LoadedSettings(settings, meta)]
        )
        # convert() error: Could not convert value
        # No value set for required option
        # Could convert loaded settings to instance
        with pytest.raises(exceptions.InvalidSettingsError, match=re.escape(err)):
            _core.convert(merged, state)

    def test_load_resolve_default_paths(self, tmp_path: Path) -> None:
        """
        Relative paths in default values can be resolved relative to a user specified
        base dir.
        """

        @settings(frozen=True)
        class S:
            p: Path = Path("tests")

        result = _core.load(S, "test")
        assert result == S(Path.cwd().joinpath("tests"))
        result = _core.load(S, "test", base_dir=tmp_path)
        assert result == S(tmp_path.joinpath("tests"))

    def test_load_settings_resolve_default_paths(self, tmp_path: Path) -> None:
        """
        Relative paths in default values can be resolved relative to a user specified
        base dir.
        """

        @settings(frozen=True)
        class S:
            p: Path = Path("tests")

        result = _core.load_settings(S, [])
        assert result == S(Path.cwd().joinpath("tests"))
        result = _core.load_settings(S, [], base_dir=tmp_path)
        assert result == S(tmp_path.joinpath("tests"))


class TestLogging:
    """
    Test emitted log messages.
    """

    def test_successfull_loading(
        self,
        caplog: pytest.LogCaptureFixture,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        In case of success, only DEBUG messages are emitted.
        """

        @settings
        class S:
            opt: str

        sf1 = tmp_path.joinpath("sf1.toml")
        sf1.write_text('[test]\nopt = "spam"\n')
        sf2 = tmp_path.joinpath("sf2.toml")
        sf2.write_text('[test]\nopt = "eggs"\n')
        monkeypatch.setenv("TEST_SETTINGS", str(sf2))
        monkeypatch.setenv("TEST_OPT", "bacon")

        caplog.set_level(logging.DEBUG)

        _core.load(S, "test", [sf1])

        assert caplog.record_tuples == [
            (
                "typed_settings",
                logging.DEBUG,
                "Env var for config files: TEST_SETTINGS",
            ),
            ("typed_settings", logging.DEBUG, f"Loading settings from: {sf1}"),
            ("typed_settings", logging.DEBUG, f"Loading settings from: {sf2}"),
            (
                "typed_settings",
                logging.DEBUG,
                "Looking for env vars with prefix: TEST_",
            ),
            ("typed_settings", logging.DEBUG, "Env var found: TEST_OPT"),
        ]

    def test_optional_files_not_found(
        self,
        caplog: pytest.LogCaptureFixture,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        Non-existing optional files emit an INFO message if file was specified
        by the app (passed to "load_settings()") an a WARNING message if the
        file was specified via an environment variable.
        """

        @settings
        class S:
            opt: str = ""

        sf1 = tmp_path.joinpath("sf1.toml")
        sf2 = tmp_path.joinpath("sf2.toml")
        monkeypatch.setenv("TEST_SETTINGS", str(sf2))

        caplog.set_level(logging.DEBUG)

        _core.load(S, "test", [sf1])

        assert caplog.record_tuples == [
            (
                "typed_settings",
                logging.DEBUG,
                "Env var for config files: TEST_SETTINGS",
            ),
            ("typed_settings", logging.INFO, f"Config file not found: {sf1}"),
            (
                "typed_settings",
                logging.WARNING,
                f"Config file from TEST_SETTINGS not found: {sf2}",
            ),
            (
                "typed_settings",
                logging.DEBUG,
                "Looking for env vars with prefix: TEST_",
            ),
            ("typed_settings", logging.DEBUG, "Env var not found: TEST_OPT"),
        ]

    def test_mandatory_files_not_found(
        self,
        caplog: pytest.LogCaptureFixture,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        In case of success, only ``debug`` messages are emitted.
        """

        @settings
        class S:
            opt: str = ""

        sf1 = tmp_path.joinpath("sf1.toml")
        monkeypatch.setenv("TEST_SETTINGS", f"!{sf1}")

        caplog.set_level(logging.DEBUG)

        with pytest.raises(FileNotFoundError):
            _core.load(S, "test")

        assert caplog.record_tuples == [
            (
                "typed_settings",
                logging.DEBUG,
                "Env var for config files: TEST_SETTINGS",
            ),
            (
                "typed_settings",
                logging.ERROR,
                f"Mandatory config file not found: {sf1}",
            ),
        ]



"""
Tests for supported CLI param types for both, Click and argparse.
"""

import re
from datetime import date, datetime, timedelta, timezone
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    FrozenSet,
    Generic,
    List,
    MutableSequence,
    MutableSet,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

import click
import click.testing
import pytest

from typed_settings import (
    SecretStr,
    cli_argparse,
    click_options,
    default_loaders,
    option,
    settings,
)
from typed_settings.types import ST


class CliResult(click.testing.Result, Generic[ST]):
    """A container for the settings passed to a test CLI."""

    settings: Optional[ST]


Cli = Callable[..., CliResult[ST]]
ArgParser = Callable[..., ST]


def make_cli(settings_cls: Type[ST]) -> Cli:
    """
    Return a function that invokes a Click CLI with Typed Settings options
    for *settings_cls*.

    That functions returns a :class:`CliResult` with the loaded settings
    instance (:attrs:`CliResult.settings`).
    """

    class Runner(click.testing.CliRunner):
        settings: Optional[ST]

        def invoke(self, *args: Any, **kwargs: Any) -> CliResult:
            result = super().invoke(*args, **kwargs)
            cli_result: CliResult[ST] = CliResult(**result.__dict__)
            try:
                cli_result.settings = self.settings
            except AttributeError:
                cli_result.settings = None
            return cli_result

    runner = Runner()

    def run(*args: Any, **kwargs: Any) -> CliResult:
        @click.group(invoke_without_command=True)
        @click_options(settings_cls, default_loaders("test"))
        def cli(settings: ST) -> None:
            runner.settings = settings

        result = runner.invoke(cli, args, catch_exceptions=False, **kwargs)
        assert result.exit_code == 0
        return result

    return run


def make_argparser(settings_cls: Type[ST]) -> ArgParser:
    """
    Create an argument parser for the argparse tests.
    """

    def parse_args(*args: str) -> ST:
        parser, ms = cli_argparse.make_parser(settings_cls, default_loaders("test"))
        namespace = parser.parse_args(args)
        return cli_argparse.namespace2settings(
            settings_cls, namespace, merged_settings=ms
        )

    return parse_args


class LeEnum(Enum):
    """A simple enum for testing."""

    spam = "Le spam"
    eggs = "Le eggs"


class ParamBase:
    """
    Base class for test parameters.

    Sublcasses must define:

    - A settings class
    - A list of expected "--help" outputs for each option
    - Optionally, required arguments for options with no default.
    - An instance of the settings class with expected default option
        values.
    - A list of Cli options and the expected settings result.
    """

    @settings
    class Settings:
        """Just a dummy settings class."""

    expected_help: ClassVar[List[str]] = []

    env_vars: ClassVar[Dict[str, str]] = {}
    expected_env_var_defaults: ClassVar[List[str]] = []

    default_options: ClassVar[List[str]] = []
    expected_defaults: ClassVar[Any] = None

    cli_options: ClassVar[List[str]] = []
    expected_settings: ClassVar[Any] = None

    classes: ClassVar[List[Type["ParamBase"]]] = []

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Register subclasses that handle the various param types.
        """
        super().__init_subclass__(**kwargs)
        cls.classes.append(cls)

    @classmethod
    def get_expected_help(cls, prefix: str) -> List[str]:
        return getattr(cls, f"{prefix}_expected_help", cls.expected_help)

    @classmethod
    def get_expected_env_var_defaults(cls, prefix: str) -> List[str]:
        return getattr(
            cls,
            f"{prefix}_expected_env_var_defaults",
            cls.expected_env_var_defaults,
        )


class TestBoolParam(ParamBase):
    """
    Test boolean flags.
    """

    @settings(kw_only=True)
    class Settings:
        a: bool
        b: bool = True
        c: bool = False
        d: Optional[bool]
        e: Optional[bool] = None

    click_expected_help = [
        "  --a / --no-a  [required]",
        "  --b / --no-b  [default: b]",
        "  --c / --no-c  [default: no-c]",
        "  --d / --no-d",
        "  --e / --no-e",
    ]
    argparse_expected_help = [
        "  --a, --no-a  [required]",
        "  --b, --no-b  [default: True]",
        "  --c, --no-c  [default: False]",
        "  --d, --no-d",
        "  --e, --no-e",
    ]

    env_vars = {"A": "1", "B": "0"}
    click_expected_env_var_defaults = [
        "  --a / --no-a  [default: a]",
        "  --b / --no-b  [default: no-b]",
        "  --c / --no-c  [default: no-c]",
        "  --d / --no-d",
        "  --e / --no-e",
    ]
    argparse_expected_env_var_defaults = [
        "  --a, --no-a  [default: True]",
        "  --b, --no-b  [default: False]",
        "  --c, --no-c  [default: False]",
        "  --d, --no-d",
        "  --e, --no-e",
    ]

    default_options = ["--a"]
    expected_defaults = Settings(a=True, b=True, c=False, d=None, e=None)

    cli_options = ["--no-a", "--no-b", "--c"]
    expected_settings = Settings(a=False, b=False, c=True, d=None, e=None)


class TestIntFloatStrParam(ParamBase):
    """
    Test int, float and str cli_options.
    """

    @settings(kw_only=True)
    class Settings:
        a: str = option(default="spam")
        b: SecretStr = option(default=SecretStr("spam"))
        c: int = 0
        d: float = 0
        # Test explicit and implicit "Optional" variants
        e: Optional[str]
        f: Union[None, int] = None
        g: Union[int, None] = 0

    click_expected_help = [
        "  --a TEXT     [default: spam]",
        "  --b TEXT     [default: (*******)]",
        "  --c INTEGER  [default: 0]",
        "  --d FLOAT    [default: 0.0]",
        "  --e TEXT",
        "  --f INTEGER",
        "  --g INTEGER  [default: 0]",
    ]
    argparse_expected_help = [
        "  --a TEXT    [default: spam]",
        "  --b TEXT    [default: (*******)]",
        "  --c INT     [default: 0]",
        "  --d FLOAT   [default: 0.0]",
        "  --e TEXT",
        "  --f INT",
        "  --g INT     [default: 0]",
    ]

    env_vars = {"A": "eggs", "B": "bacon", "C": "42", "D": "3.14"}
    click_expected_env_var_defaults = [
        "  --a TEXT     [default: eggs]",
        "  --b TEXT     [default: (*******)]",
        "  --c INTEGER  [default: 42]",
        "  --d FLOAT    [default: 3.14]",
        "  --e TEXT",
        "  --f INTEGER",
        "  --g INTEGER  [default: 0]",
    ]
    argparse_expected_env_var_defaults = [
        "  --a TEXT    [default: eggs]",
        "  --b TEXT    [default: (*******)]",
        "  --c INT     [default: 42]",
        "  --d FLOAT   [default: 3.14]",
        "  --e TEXT",
        "  --f INT",
        "  --g INT     [default: 0]",
    ]

    expected_defaults = Settings(e=None)

    cli_options = ["--a=eggs", "--b=pwd", "--c=3", "--d=3.1"]
    expected_settings = Settings(a="eggs", b=SecretStr("pwd"), c=3, d=3.1, e=None)


class TestDateTimeParam(ParamBase):
    """
    Test datetime cli_options.
    """

    @settings
    class Settings:
        a: datetime = datetime.fromtimestamp(0, timezone.utc)
        b: datetime = datetime.fromtimestamp(0, timezone.utc)
        c: datetime = datetime.fromtimestamp(0, timezone.utc)
        d: Optional[datetime] = None

    click_expected_help = [
        "  --a [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
        "                                  [default: 1970-01-01T00:00:00+00:00]",
        "  --b [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
        "                                  [default: 1970-01-01T00:00:00+00:00]",
        "  --c [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
        "                                  [default: 1970-01-01T00:00:00+00:00]",
        "  --d [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
    ]
    argparse_expected_help = [
        "  --a YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
        "                        [default: 1970-01-01T00:00:00+00:00]",
        "  --b YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
        "                        [default: 1970-01-01T00:00:00+00:00]",
        "  --c YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
        "                        [default: 1970-01-01T00:00:00+00:00]",
        "  --d YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
    ]

    env_vars = {
        "A": "2021-05-04T13:37:00Z",
        "B": "2021-05-04T13:37:00",
        "C": "2021-05-04",
    }
    click_expected_env_var_defaults = [
        "  --a [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
        "                                  [default: 2021-05-04T13:37:00+00:00]",
        "  --b [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
        "                                  [default: 2021-05-04T13:37:00]",
        "  --c [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
        "                                  [default: 2021-05-04T00:00:00]",
        "  --d [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
    ]
    argparse_expected_env_var_defaults = [
        "  --a YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
        "                        [default: 2021-05-04T13:37:00+00:00]",
        "  --b YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
        "                        [default: 2021-05-04T13:37:00]",
        "  --c YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
        "                        [default: 2021-05-04T00:00:00]",
        "  --d YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
    ]

    expected_defaults = Settings()

    cli_options = [
        "--a=2020-05-04",
        "--b=2020-05-04T13:37:00",
        "--c=2020-05-04T13:37:00+00:00",
    ]
    expected_settings = Settings(
        datetime(2020, 5, 4),
        datetime(2020, 5, 4, 13, 37),
        datetime(2020, 5, 4, 13, 37, tzinfo=timezone.utc),
    )


class TestDateParam(ParamBase):
    """
    Test date cli_options.
    """

    @settings
    class Settings:
        a: date
        b: date = date(1970, 1, 1)
        c: Optional[date] = None

    click_expected_help = [
        "  --a [%Y-%m-%d]  [required]",
        "  --b [%Y-%m-%d]  [default: 1970-01-01]",
        "  --c [%Y-%m-%d]",
    ]
    argparse_expected_help = [
        "  --a YYYY-MM-DD  [required]",
        "  --b YYYY-MM-DD  [default: 1970-01-01]",
        "  --c YYYY-MM-DD",
    ]

    env_vars = {
        "A": "2021-05-04",
    }
    click_expected_env_var_defaults = [
        "  --a [%Y-%m-%d]  [default: 2021-05-04]",
        "  --b [%Y-%m-%d]  [default: 1970-01-01]",
        "  --c [%Y-%m-%d]",
    ]
    argparse_expected_env_var_defaults = [
        "  --a YYYY-MM-DD  [default: 2021-05-04]",
        "  --b YYYY-MM-DD  [default: 1970-01-01]",
        "  --c YYYY-MM-DD",
    ]

    default_options = ["--a=2020-05-04"]
    expected_defaults = Settings(date(2020, 5, 4))

    cli_options = [
        "--a=2020-05-04",
    ]
    expected_settings = Settings(date(2020, 5, 4))


class TestTimedeltaParam(ParamBase):
    """
    Test date cli_options.
    """

    @settings
    class Settings:
        a: timedelta
        b: timedelta = timedelta(days=1, seconds=4)
        c: timedelta = timedelta(days=1, seconds=4)
        d: Optional[timedelta] = None

    click_expected_help = [
        "  --a [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                                  [required]",
        "  --b [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                                  [default: 1d4s]",
        "  --c [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                                  [default: 1d4s]",
        "  --d [-][Dd][HHh][MMm][SS[.ffffff]s]",
    ]
    argparse_expected_help = [
        "  --a [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                        [required]",
        "  --b [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                        [default: 1d4s]",
        "  --c [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                        [default: 1d4s]",
        "  --d [-][Dd][HHh][MMm][SS[.ffffff]s]",
    ]

    env_vars = {
        "A": "P1DT03H04M",
        "B": "3h4m",
        "C": "03:04:02",
    }
    click_expected_env_var_defaults = [
        "  --a [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                                  [default: 1d3h4m]",
        "  --b [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                                  [default: 3h4m]",
        "  --c [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                                  [default: 3h4m2s]",
        "  --d [-][Dd][HHh][MMm][SS[.ffffff]s]",
    ]
    argparse_expected_env_var_defaults = [
        "  --a [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                        [default: 1d3h4m]",
        "  --b [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                        [default: 3h4m]",
        "  --c [-][Dd][HHh][MMm][SS[.ffffff]s]",
        "                        [default: 3h4m2s]",
        "  --d [-][Dd][HHh][MMm][SS[.ffffff]s]",
    ]

    default_options = ["--a=1d"]
    expected_defaults = Settings(timedelta(days=1))

    cli_options = [
        "--a=1d",
        "--b=1h",
        "--c=1d2h3m4s",
    ]
    expected_settings = Settings(
        timedelta(days=1),
        timedelta(hours=1),
        timedelta(days=1, hours=2, minutes=3, seconds=4),
    )


class TestEnumParam(ParamBase):
    """
    Test enum cli_options.
    """

    @settings
    class Settings:
        a: LeEnum
        b: Optional[LeEnum]
        c: LeEnum = LeEnum.spam

    click_expected_help = [
        "  --a [spam|eggs]  [required]",
        "  --b [spam|eggs]",
        "  --c [spam|eggs]  [default: spam]",
    ]
    argparse_expected_help = [
        "  --a {spam,eggs}  [required]",
        "  --b {spam,eggs}",
        "  --c {spam,eggs}  [default: spam]",
    ]

    env_vars = {"A": "spam", "C": "eggs"}
    click_expected_env_var_defaults = [
        "  --a [spam|eggs]  [default: spam]",
        "  --b [spam|eggs]",
        "  --c [spam|eggs]  [default: eggs]",
    ]
    argparse_expected_env_var_defaults = [
        "  --a {spam,eggs}  [default: spam]",
        "  --b {spam,eggs}",
        "  --c {spam,eggs}  [default: eggs]",
    ]

    default_options = ["--a=spam"]
    expected_defaults = Settings(a=LeEnum.spam, b=None)

    cli_options = ["--a=spam", "--c=eggs"]
    expected_settings = Settings(LeEnum.spam, None, LeEnum.eggs)


class TestPathParam(ParamBase):
    """
    Test Path cli_options.
    """

    @settings
    class Settings:
        a: Path
        b: Path = Path("/")
        c: Optional[Path] = None

    click_expected_help = [
        "  --a PATH  [required]",
        "  --b PATH  [default: /]",
        "  --c PATH",
    ]
    argparse_expected_help = [
        "  --a PATH    [required]",
        "  --b PATH    [default: /]",
        "  --c PATH",
    ]

    env_vars = {"A": "/spam/eggs"}
    click_expected_env_var_defaults = [
        "  --a PATH  [default: /spam/eggs]",
        "  --b PATH  [default: /]",
        "  --c PATH",
    ]
    argparse_expected_env_var_defaults = [
        "  --a PATH    [default: /spam/eggs]",
        "  --b PATH    [default: /]",
        "  --c PATH",
    ]

    default_options = ["--a=/"]
    expected_defaults = Settings(Path("/"))

    cli_options = ["--a=/spam"]
    expected_settings = Settings(Path("/spam"))


class TestNestedParam(ParamBase):
    """
    Test cli_options for nested settings.
    """

    @settings
    class Settings:
        @settings
        class Nested:
            a: str = "nested"
            b: int = 0

        n: Nested = Nested()

    click_expected_help = [
        "  --n-a TEXT     [default: nested]",
        "  --n-b INTEGER  [default: 0]",
    ]
    argparse_expected_help = [
        "  --n-a TEXT  [default: nested]",
        "  --n-b INT   [default: 0]",
    ]

    env_vars = {"N_A": "spam", "N_B": "42"}
    click_expected_env_var_defaults = [
        "  --n-a TEXT     [default: spam]",
        "  --n-b INTEGER  [default: 42]",
    ]
    argparse_expected_env_var_defaults = [
        "  --n-a TEXT  [default: spam]",
        "  --n-b INT   [default: 42]",
    ]

    expected_defaults = Settings()

    cli_options = ["--n-a=eggs", "--n-b=3"]
    expected_settings = Settings(Settings.Nested("eggs", 3))


class TestListParam(ParamBase):
    """
    Lists (and friends) use "multiple=True".
    """

    @settings
    class Settings:
        a: List[int]
        b: Optional[List[int]]
        c: Optional[List[int]] = None
        d: Optional[List[int]] = []
        e: Sequence[datetime] = [datetime(2020, 5, 4)]
        f: MutableSequence[int] = []
        g: Set[int] = set()
        h: MutableSet[int] = set()
        i: FrozenSet[int] = frozenset()

    click_expected_help = [
        "  --a INTEGER",
        "  --b INTEGER",
        "  --c INTEGER",
        "  --d INTEGER",
        "  --e [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
        "                                  [default: 2020-05-04T00:00:00]",
        "  --f INTEGER",
        "  --g INTEGER",
        "  --h INTEGER",
        "  --i INTEGER",
    ]
    argparse_expected_help = [
        "  --a INT               [default: []]",
        "  --b INT               [default: []]",
        "  --c INT               [default: []]",
        "  --d INT               [default: []]",
        "  --e YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
        "                        [default: ['2020-05-04T00:00:00']]",
        "  --f INT               [default: []]",
        "  --g INT               [default: []]",
        "  --h INT               [default: []]",
        "  --i INT               [default: []]",
    ]

    env_vars = {
        "A": "1:2",
        "E": "2021-01-01:2021-01-02",  # Dates with times wont work here!
    }
    click_expected_env_var_defaults = [
        "  --a INTEGER                     [default: 1, 2]",
        "  --b INTEGER",
        "  --c INTEGER",
        "  --d INTEGER",
        "  --e [%Y-%m-%d|%Y-%m-%dT%H:%M:%S|%Y-%m-%dT%H:%M:%S%z]",
        "                                  [default: 2021-01-01T00:00:00,",
        "                                  2021-01-02T00:00:00]",
        "  --f INTEGER",
        "  --g INTEGER",
        "  --h INTEGER",
        "  --i INTEGER",
    ]
    argparse_expected_env_var_defaults = [
        "  --a INT               [default: [1, 2]]",
        "  --b INT               [default: []]",
        "  --c INT               [default: []]",
        "  --d INT               [default: []]",
        "  --e YYYY-MM-DD[Thh:mm:ss[+xx:yy]]",
        "                        [default: ['2021-01-01T00:00:00',",
        "                        '2021-01-02T00:00:00']]",
        "  --f INT               [default: []]",
        "  --g INT               [default: []]",
        "  --h INT               [default: []]",
        "  --i INT               [default: []]",
    ]

    default_options = ["--a=1"]
    # click always gives us an empty list, never "None"
    expected_defaults = Settings(a=[1], b=[], c=[])

    cli_options = [
        "--a=1",
        "--a=2",
        "--e=2020-01-01",
        "--e=2020-01-02",
        "--f=3",
        "--g=4",
        "--h=5",
        "--i=6",
    ]
    expected_settings = Settings(
        [1, 2],
        [],  # click always gives us an empty list, never "None"
        [],  # click always gives us an empty list, never "None"
        [],
        [datetime(2020, 1, 1), datetime(2020, 1, 2)],
        [3],
        {4},
        {5},
        frozenset({6}),
    )


class TestTupleParam(ParamBase):
    """
    Tuples are handled either like the list variant with multiple=True or
    like the struct variant with nargs=x.
    """

    @settings
    class Settings:
        a: Tuple[int, ...] = (0,)
        b: Tuple[int, float, str] = (0, 0.0, "")
        c: Optional[Tuple[int, float, str]] = None

    click_expected_help = [
        "  --a INTEGER                  [default: 0]",
        "  --b <INTEGER FLOAT TEXT>...  [default: 0, 0.0, ]",
        "  --c <INTEGER FLOAT TEXT>...",
    ]
    argparse_expected_help = [
        "  --a INT             [default: [0]]",
        "  --b INT FLOAT TEXT  [default: (0, 0.0, '')]",
        "  --c INT FLOAT TEXT",
    ]

    env_vars = {"A": "1:2", "B": "42:3.14:spam"}
    click_expected_env_var_defaults = [
        "  --a INTEGER                  [default: 1, 2]",
        "  --b <INTEGER FLOAT TEXT>...  [default: 42, 3.14, spam]",
        "  --c <INTEGER FLOAT TEXT>...",
    ]
    argparse_expected_env_var_defaults = [
        "  --a INT             [default: [1, 2]]",
        "  --b INT FLOAT TEXT  [default: (42, 3.14, 'spam')]",
        "  --c INT FLOAT TEXT",
    ]

    expected_defaults = Settings()

    cli_options = ["--a=1", "--a=2", "--b", "1", "2.3", "spam"]
    expected_settings = Settings((1, 2), (1, 2.3, "spam"))

    def test_wrong_default_length(self) -> None:
        """
        Default values for tuples must have the exact shape defined in by
        their type.

        Too long tuples are automatically truncated (by attrs or click),
        but too short default values are an error.
        """

        @settings
        class Settings:
            a: Tuple[int, int, int] = (0, 1)  # type: ignore

        run = make_cli(Settings)

        p = re.escape(
            "Invalid default (0, 1) for option 'a' with type "
            "typing.Tuple[int, int, int]: IterableValidationError"
        )
        with pytest.raises(ValueError, match=f"{p}.*"):
            run("--help")


class TestNestedTupleParam(ParamBase):
    """
    Lists of tuples use "multiple=True" and "nargs=x".
    """

    @settings
    class Settings:
        a: List[Tuple[int, int]] = option(factory=list)

    click_expected_help = [
        "  --a <INTEGER INTEGER>...",
    ]
    argparse_expected_help = [
        "  --a INT INT  [default: []]",
    ]

    # A list of tuples cannot be loaded with the default converter,
    # so we skip this test here.
    env_vars: ClassVar[Dict[str, Any]] = {}
    click_expected_env_var_defaults = [
        "  --a <INTEGER INTEGER>...",
    ]
    argparse_expected_env_var_defaults = [
        "  --a INT INT  [default: []]",
    ]

    expected_defaults = Settings()

    cli_options = ["--a", "1", "2", "--a", "3", "4"]
    expected_settings = Settings([(1, 2), (3, 4)])


class TestDictParam(ParamBase):
    """
    Dict params use "="-spearated key-value pairs and "multiple=True".
    """

    @settings
    class Settings:
        a: Dict[str, str]  # Test "None" value
        b: Dict[str, str] = {}  # Test empty value
        c: Dict[str, str] = {"default": "value"}
        d: Optional[Dict[str, str]] = None

    click_expected_help = [
        "  --a KEY=VALUE",
        "  --b KEY=VALUE",
        "  --c KEY=VALUE  [default: default=value]",
        "  --d KEY=VALUE",
    ]
    argparse_expected_help = [
        "  --a KEY=VALUE  [default: ]",
        "  --b KEY=VALUE  [default: ]",
        "  --c KEY=VALUE  [default: default=value]",
        "  --d KEY=VALUE  [default: ]",
    ]

    # A dictionary cannot be loaded with the default converter,
    # so we skip this test here.
    env_vars: ClassVar[Dict[str, Any]] = {}
    click_expected_env_var_defaults = [
        "  --a KEY=VALUE",
        "  --b KEY=VALUE",
        "  --c KEY=VALUE  [default: default=value]",
        "  --d KEY=VALUE",
    ]
    argparse_expected_env_var_defaults = [
        "  --a KEY=VALUE  [default: ]",
        "  --b KEY=VALUE  [default: ]",
        "  --c KEY=VALUE  [default: default=value]",
        "  --d KEY=VALUE  [default: ]",
    ]

    default_options = ["--a=k=v"]
    expected_defaults = Settings({"k": "v"}, d={})  # "d" is never None

    cli_options = [
        "--a",
        "key0=val0",
        "--c",
        "key1=val1",
        "--c",
        "key-2=val-2",
        "--c",
        "key 3=oi oi",
    ]
    expected_settings = Settings(
        {"key0": "val0"},
        {},
        {"key1": "val1", "key-2": "val-2", "key 3": "oi oi"},
        {},
    )


class TestNoTypeParam(ParamBase):
    """
    Test option without type annotation.
    """

    @settings
    class Settings:
        a = option(default="spam")  # type: ignore

    click_expected_help = [
        "  --a TEXT  [default: spam]",
    ]
    argparse_expected_help = [
        "  --a A       [default: spam]",
    ]

    env_vars = {"A": "eggs"}
    click_expected_env_var_defaults = [
        "  --a TEXT  [default: eggs]",
    ]
    argparse_expected_env_var_defaults = [
        "  --a A       [default: eggs]",
    ]

    expected_defaults = Settings("spam")  # type: ignore

    cli_options = ["--a=eggs"]
    expected_settings = Settings(a="eggs")  # type: ignore


class TestAttrsDefaultFactory(ParamBase):
    """
    Test params with ``attrs`` default factories.
    """

    @settings
    class Settings:
        a: str = option(factory=lambda: "spam")

    click_expected_help = [
        "  --a TEXT  [default: (dynamic)]",
    ]
    argparse_expected_help = [
        "  --a TEXT    [default: (dynamic)]",
    ]

    env_vars = {"A": "eggs"}
    click_expected_env_var_defaults = [
        "  --a TEXT  [default: eggs]",
    ]
    argparse_expected_env_var_defaults = [
        "  --a TEXT    [default: eggs]",
    ]

    default_options = []
    expected_defaults = Settings("spam")

    cli_options = ["--a=bacon"]
    expected_settings = Settings("bacon")


@pytest.fixture(params=ParamBase.classes)
def param_cls(request: pytest.FixtureRequest) -> Type[ParamBase]:
    """
    Create a test for each "ParamBase" sublclass and return that subclass.
    """
    return cast(Type[ParamBase], request.param)


class TestClick:
    """
    Test the supported types with Click.
    """

    prefix = "click"

    @pytest.fixture
    def cli(self, param_cls: Type[ParamBase]) -> Cli:
        return make_cli(param_cls.Settings)

    def test_help(self, param_cls: Type[ParamBase], cli: Cli[ST]) -> None:
        """
        The genereated CLI has a proper help output.
        """
        result = cli("--help")

        assert result.output.splitlines()[:-1] == [
            "Usage: cli [OPTIONS] COMMAND [ARGS]...",
            "",
            "Options:",
            *param_cls.get_expected_help(self.prefix),
        ]

    def test_defaults_from_envvars(
        self,
        param_cls: Type[ParamBase],
        cli: Cli[ST],
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        Previously loaded settings (e.g., from env vars) can be converted
        to click options.

        Regression test for #11
        """
        for var, val in param_cls.env_vars.items():
            monkeypatch.setenv(f"TEST_{var}", val)

        result = cli("--help")

        assert result.output.splitlines()[:-1] == [
            "Usage: cli [OPTIONS] COMMAND [ARGS]...",
            "",
            "Options:",
            *param_cls.get_expected_env_var_defaults(self.prefix),
        ]

    def test_defaults(self, param_cls: Type[ParamBase], cli: Cli[ST]) -> None:
        """
        Arguments of the generated CLI have default values.
        """
        result = cli(*param_cls.default_options)
        assert result.output == ""
        assert result.settings == param_cls.expected_defaults

    def test_options(self, param_cls: Type[ParamBase], cli: Cli[ST]) -> None:
        """
        Default values can be overriden by passing the corresponding args.
        """
        result = cli(*param_cls.cli_options)
        assert result.output == ""
        assert result.settings == param_cls.expected_settings


class TestArgparse:
    """
    Test the supported types with argparse.
    """

    prefix = "argparse"

    @pytest.fixture
    def arg_parser(self, param_cls: Type[ParamBase]) -> ArgParser:
        return make_argparser(param_cls.Settings)

    def test_help(
        self,
        param_cls: Type[ParamBase],
        arg_parser: ArgParser,
        capsys: pytest.CaptureFixture,
    ) -> None:
        """
        The genereated CLI has a proper help output.
        """
        with pytest.raises(SystemExit):
            arg_parser("--help")
        out, err = capsys.readouterr()
        print(out)
        assert err == ""

        lines: List[str] = []
        for line in out.splitlines():
            if line.startswith("  --") or lines:
                lines.append(line)
        assert lines == param_cls.get_expected_help(self.prefix)

    def test_defaults_from_envvars(
        self,
        param_cls: Type[ParamBase],
        arg_parser: ArgParser,
        capsys: pytest.CaptureFixture,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        Previously loaded settings (e.g., from env vars) can be converted
        to click options.

        Regression test for #11
        """
        for var, val in param_cls.env_vars.items():
            monkeypatch.setenv(f"TEST_{var}", val)

        with pytest.raises(SystemExit):
            arg_parser("--help")
        out, err = capsys.readouterr()
        print(out)

        lines: List[str] = []
        for line in out.splitlines():
            if line.startswith("  --") or lines:
                lines.append(line)
        assert lines == param_cls.get_expected_env_var_defaults(self.prefix)

    def test_defaults(
        self,
        param_cls: Type[ParamBase],
        arg_parser: ArgParser,
    ) -> None:
        """
        Arguments of the generated CLI have default values.
        """
        result = arg_parser(*param_cls.default_options)
        assert result == param_cls.expected_defaults

    def test_options(
        self,
        param_cls: Type[ParamBase],
        arg_parser: ArgParser,
    ) -> None:
        """
        Default values can be overriden by passing the corresponding args.
        """
        result = arg_parser(*param_cls.cli_options)
        assert result == param_cls.expected_settings



"""
Test behavior when no optional dependencies are installed.
"""

import sys

import pytest

import typed_settings
import typed_settings.loaders

from . import conftest


try:
    import attrs, click  # noqa: E401, F401, I001

    pytestmark = pytest.mark.skip(reason="Optional dependencies are installed")
except ImportError:
    pass


@pytest.mark.parametrize(
    "dep, name",
    [
        ("attrs", "combine"),
        ("attrs", "evolve"),
        ("attrs", "option"),
        ("attrs", "secret"),
        ("attrs", "settings"),
        ("click", "click_options"),
        ("click", "pass_settings"),
    ],
)
def test_import_error(dep: str, name: str) -> None:
    """
    An ImportError is raised when trying to get an attrib from "typed_settings" that
    requires an optional import.
    """
    with pytest.raises(ImportError) as exc_info:
        getattr(typed_settings, name)
    assert dep in str(exc_info.value)


def test_attribute_not_found() -> None:
    """
    Names not provided by typed settings lead to an attribute error as expected.
    """
    with pytest.raises(AttributeError):
        _ = typed_settings.spam


def test_load(
    settings_clss: conftest.SettingsClasses, monkeypatch: pytest.MonkeyPatch
) -> None:
    """
    "load()" works with dataclasses when no optional dependencies are installed.
    """
    monkeypatch.setenv("TEST_HOST_NAME", "n")
    monkeypatch.setenv("TEST_HOST_PORT", "1")
    monkeypatch.setenv("TEST_URL", "u")
    monkeypatch.setenv("TEST_DEFAULT", "2")
    result: conftest.Settings = typed_settings.load(settings_clss[0], "test")
    assert result == conftest.Settings(conftest.Host("n", 1), "u", 2)


def test_load_settings(settings_clss: conftest.SettingsClasses) -> None:
    """
    "load_settings()" and "default_converter()" work with dataclasses when no optional
    dependencies are installed.
    """
    default = conftest.Settings(conftest.Host("n", 1), "u", 2)
    result: conftest.Settings = typed_settings.load_settings(
        settings_clss[0],
        loaders=[typed_settings.loaders.InstanceLoader(default)],
        converter=typed_settings.default_converter(),
    )
    assert result == default


def test_cli(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Basic test "cli()" - simple CLI for a simple settings class.
    """
    monkeypatch.setattr(
        sys, "argv", ["cli", "--host-name=n", "--host-port=1", "--url=u", "--default=2"]
    )

    @typed_settings.cli(conftest.Settings, "test")
    def cli(settings: conftest.Settings) -> None:
        assert settings == conftest.Settings(conftest.Host("n", 1), "u", 2)

    cli()



"""
Internal data structures.
"""

import dataclasses
from collections.abc import Collection
from enum import Enum
from pathlib import Path
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    NamedTuple,
    NewType,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from ._compat import PY_39, PY_310
from .constants import SECRET_REPR


if TYPE_CHECKING:
    from typing_extensions import TypeGuard

__all__ = [
    "AUTO",
    "SECRET_REPR",
    "T",
    "ET",
    "ST",
    "SettingsClass",
    "SettingsInstance",
    "SettingsDict",
    "OptionName",
    "OptionPath",
    "OptionInfo",
    "OptionList",
    "OptionDict",
    "LoaderMeta",
    "LoadedValue",
    "LoadedSettings",
    "MergedSettings",
    "Secret",
    "SecretStr",
    "SECRETS_TYPES",
    "is_new_type",
]


#: A generic TypeVar
T = TypeVar("T")
#: A TypeVar for :class:`~enum.Enum` types
ET = TypeVar("ET", bound=Enum)  # Enum type
#: A TypeVar for settings instances
ST = TypeVar("ST")  # Type var for SettingsInstance

SettingsClass = type
SettingsInstance = Any
OptionName = str
OptionPath = str
SettingsDict = Dict[OptionName, Union[Any, "SettingsDict"]]
"""
A dictionary with all loaded settings.

Values are not converted to their final type yet.
"""


class _Auto:
    """
    Sentinel class to indicate the lack of a value when ``None`` is ambiguous.

    ``_Auto`` is a singleton. There is only ever one of it.
    """

    _singleton = None

    def __new__(cls) -> "_Auto":
        if _Auto._singleton is None:
            _Auto._singleton = super().__new__(cls)
        return _Auto._singleton

    def __repr__(self) -> str:
        return "AUTO"


AUTO = _Auto()
"""
Sentinel to indicate the lack of a value when ``None`` is ambiguous.
"""


def _type2name(value: Union[str, Any]) -> str:
    """
    Return either *value* if it is a str or else its type name.
    """
    if isinstance(value, str):
        return value
    return type(value).__name__


@dataclasses.dataclass(frozen=True)
class OptionInfo:
    """
    Information about (possibly nested) option attributes.

    Each instance represents a single attribute of an apps's settings class.
    """

    parent_cls: type
    """
    The option's settings class.  This is either the root settings class or a nested
    one.
    """

    path: OptionPath
    """
    Dotted path to the option name relative to the root settings class.
    """

    name: str = dataclasses.field(init=False)

    cls: type
    default: Any
    has_no_default: bool
    default_is_factory: bool

    is_secret: bool = False
    converter: Optional[Callable[[Any], Any]] = None
    metadata: Dict[Any, Any] = dataclasses.field(default_factory=dict)

    @property
    def has_default(self) -> bool:
        return not self.has_no_default

    def __post_init__(self) -> None:
        _prefix, _, name = self.path.rpartition(".")
        object.__setattr__(self, "name", name)


OptionList = Tuple[OptionInfo, ...]
"""
A flat list of all available options, including those from nested settings.
"""

if PY_39:
    OptionDict = MappingProxyType[OptionPath, OptionInfo]
else:
    from typing import Mapping

    OptionDict = Mapping[OptionPath, OptionInfo]  # type: ignore
"""
A dict version of :class:`OptionList`.
"""


class LoaderMeta:
    """
    Meta data about the loader that loaded a set of option values.

    That data is used to improve error reporting and for converting loaded values to the
    target type.
    """

    def __init__(self, name: Union[str, Any], base_dir: Optional[Path] = None):
        self._name: str = _type2name(name)
        self._base_dir = base_dir or Path.cwd()

    def __str__(self) -> str:
        return f"{type(self).__name__}({self._name!r}, {self.base_dir!r})"

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, type(self))
            and self.name == other.name
            and self.base_dir == other.base_dir
        )

    @property
    def name(self) -> str:
        """
        The loader's name.  Can be a string or the loader class itself (it's class name
        is will then be used).
        """
        return self._name

    @property
    def base_dir(self) -> Path:
        """
        The loader's base directory.

        It is used to resolve relative paths in loaded option values in the proper
        context.

        For most loaders, this should be the *cwd* (which is also the default).  For
        file loaders, the parent directory of a loaded file might be a better
        alternative.
        """
        return self._base_dir


class LoadedValue(NamedTuple):
    """
    A container for a loaded option value and the meta data of the originating loader.
    """

    value: Any
    """
    The loaded option value.
    """

    loader_meta: LoaderMeta
    """
    Meta data of the loader that loaded the corresponding value.
    """


@dataclasses.dataclass(frozen=True)
class LoadedSettings:
    """
    A container for the settings loaded by a single loader, and the meta data of that
    loader.
    """

    settings: SettingsDict
    """
    The loaded settings values.
    """

    meta: LoaderMeta
    """
    Meta data of the loader that loaded the settings.
    """


MergedSettings = Dict[OptionPath, LoadedValue]
"""
A dict that maps a dotted option path to a loaded option value.

The values may come from different loaders, so each option values stores the meta data
of it's loader.
"""


class SecretStr(str):
    """
    A subclass of :class:`str` that masks the output of :func:`repr()`.

    It is less secure than :class:`Secret` but is a drop-in replacement for normal
    strings.

    The main use case is avoiding accidental secrets leakage via tracebacks.
    It also helps to enforce secret usage via Typing.

    It does **not help** when you:

    - :func:`print()` it
    - :class:`str` it
    - Log it
    - Use it in an f-string (``f"{val}"``)

    .. versionadded:: 2.0.0
    """

    def __repr__(self) -> str:
        """
        Return a secret representation if a non-empty value is set, else a repr for an
        empty string.
        """
        return f"{SECRET_REPR!r}" if self else "''"


class Secret(Generic[T]):
    """
    A secret wrapper around any value.

    It makes it very hard to accidentally leak the secret, even when printing or logging
    it.

    You need to explicitly call :meth:`get_secret_value()` to get the wrapped value.
    Thus, it is no drop-in replacement for the wrapped data.

    See :class:`SecretStr` if you need a drop-in replacement for strings, even if it is
    not quite as safe.

    You can use :class:`bool` to get the boolean value of the wrapped secret.  Other
    protocol methods (e.g., for length or comparison operators) are not implemented.

    .. versionadded:: 2.0.0
    """

    def __init__(self, secret_value: T) -> None:
        self._is_collection = isinstance(secret_value, Collection)
        self._secret_value = secret_value

    def __bool__(self) -> bool:
        """
        Return the boolean representation of the stored secret.
        """
        return bool(self._secret_value)

    def __repr__(self) -> str:
        """
        Return a secret representation if a non-empty value is set, else a repr for an
        empty string.
        """
        r = repr(
            self._secret_value
            if not self._secret_value and self._is_collection
            else SECRET_REPR
        )
        return f"{self.__class__.__name__}({r})"

    def __str__(self) -> str:
        """
        Return a secret representation if a non-empty value is set, else an empty
        string.
        """
        return str(
            self._secret_value
            if not self._secret_value and self._is_collection
            else SECRET_REPR
        )

    def get_secret_value(self) -> T:
        """
        Return the wrapped secret value.
        """
        return self._secret_value


class SecretRepr:
    def __call__(self, v: Any) -> str:
        return repr(v if not v and isinstance(v, Collection) else SECRET_REPR)

    def __repr__(self) -> str:
        return "***"


class NewTypeLike(Protocol):
    __supertype__: Type


def is_new_type(obj: Any) -> "TypeGuard[NewTypeLike]":
    """
    Return ``True`` if *obj* is a :class:`~typing.NewType`.
    """
    if PY_310:
        return isinstance(obj, NewType)
    else:
        return hasattr(obj, "__supertype__") and isinstance(obj.__supertype__, type)


SECRETS_TYPES = (Secret, SecretStr)
"""
Types that mask the repr of their values.
"""
